"""Tests the `mistral._utils.convert_message_params` function."""

import pytest
from mistralai.models import (
    AssistantMessage,
    FunctionCall,
    ImageURL,
    ImageURLChunk,
    SystemMessage,
    TextChunk,
    ToolCall,
    ToolMessage,
    UserMessage,
)
from mistralai.types.basemodel import Unset

from mirascope.core.base import (
    AudioPart,
    BaseMessageParam,
    ImagePart,
    ImageURLPart,
    TextPart,
    ToolCallPart,
    ToolResultPart,
)
from mirascope.core.mistral._utils._convert_message_params import convert_message_params


def test_convert_message_params() -> None:
    """Tests the `convert_message_params` function."""

    message_params: list[
        BaseMessageParam | AssistantMessage | SystemMessage | ToolMessage | UserMessage
    ] = [
        UserMessage(content="Hello"),
        BaseMessageParam(role="user", content="Hello"),
        BaseMessageParam(role="user", content=[TextPart(type="text", text="Hello")]),
        BaseMessageParam(role="assistant", content="Hello"),
        BaseMessageParam(role="system", content="Hello"),
        BaseMessageParam(role="tool", content="Hello"),
        AssistantMessage(
            content="Hello", tool_calls=Unset(), prefix=False, role="assistant"
        ),
        SystemMessage(content="Hello", role="system"),
        ToolMessage(content="Hello", tool_call_id=Unset(), name=Unset(), role="tool"),
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="Hello"),
                ToolResultPart(
                    type="tool_result", id="tool_id", content="result", name="tool_name"
                ),
                ImagePart(
                    type="image", media_type="image/jpeg", image=b"image", detail="auto"
                ),
                ToolCallPart(type="tool_call", name="tool_name", args={"arg": "val"}),
                ImageURLPart(
                    type="image_url", url="http://example.com/img", detail="ignored"
                ),
            ],
        ),
    ]
    converted_message_params = convert_message_params(message_params)
    assert converted_message_params == [
        UserMessage(content="Hello", role="user"),
        UserMessage(content="Hello", role="user"),
        UserMessage(content=[TextChunk(text="Hello", TYPE="text")], role="user"),
        AssistantMessage(
            content="Hello", tool_calls=Unset(), prefix=False, role="assistant"
        ),
        SystemMessage(content="Hello", role="system"),
        ToolMessage(content="Hello", tool_call_id=Unset(), name=Unset(), role="tool"),
        AssistantMessage(
            content="Hello", tool_calls=Unset(), prefix=False, role="assistant"
        ),
        SystemMessage(content="Hello", role="system"),
        ToolMessage(content="Hello", tool_call_id=Unset(), name=Unset(), role="tool"),
        UserMessage(
            content=[
                TextChunk(text="Hello", TYPE="text"),
            ],
            role="user",
        ),
        ToolMessage(
            content="result", tool_call_id="tool_id", name="tool_name", role="tool"
        ),
        AssistantMessage(
            content=[
                ImageURLChunk(
                    image_url=ImageURL(
                        url="data:image/jpeg;base64,aW1hZ2U=", detail="auto"
                    )
                )
            ],
            tool_calls=[
                ToolCall(
                    function=FunctionCall(name="tool_name", arguments={"arg": "val"}),
                    id=None,
                    type="function",
                )
            ],
            prefix=False,
            role="assistant",
        ),
        UserMessage(
            content=[
                ImageURLChunk(image_url="http://example.com/img", TYPE="image_url")
            ],
            role="user",
        ),
    ]

    with pytest.raises(
        ValueError,
        match="Mistral currently only supports text and image parts. Part provided: audio",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        AudioPart(
                            type="audio",
                            media_type="audio/wav",
                            audio=b"audio",
                        )
                    ],
                ),
            ]
        )

    with pytest.raises(
        ValueError,
        match="Invalid role: invalid_role",
    ):
        convert_message_params(
            [
                BaseMessageParam(role="invalid_role", content="Hello"),
            ]
        )

    with pytest.raises(
        ValueError,
        match="Unsupported image media type: image/svg."
        " Mistral currently only supports JPEG, PNG, GIF, and WebP images.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        ImagePart(
                            type="image",
                            media_type="image/svg",
                            image=b"image",
                            detail="auto",
                        )
                    ],
                )
            ]
        )


def test_tool_call_no_content() -> None:
    """Tests conversion of a tool_call part when there is no preceding content."""
    message = BaseMessageParam(
        role="user",
        content=[
            ToolCallPart(
                type="tool_call", name="my_tool", args={"param": "value"}, id="tc123"
            )
        ],
    )
    result = convert_message_params([message])
    expected = AssistantMessage(
        content=None,
        tool_calls=[
            ToolCall(
                function=FunctionCall(name="my_tool", arguments={"param": "value"}),
                id="tc123",
                type="function",
            )
        ],
        prefix=False,
        role="assistant",
    )
    assert result == [expected]



"""Tests the `gemini.tool` module."""

from enum import Enum
from typing import Literal

import pytest
from google.ai.generativelanguage import FunctionCall
from google.generativeai.types import (  # type: ignore
    FunctionDeclaration,
    Tool,
)

from mirascope.core.base.tool import BaseTool
from mirascope.core.gemini.tool import GeminiTool


class FormatBook(GeminiTool):
    """Returns the title and author nicely formatted."""

    title: str
    author: str
    genre: Literal["fantasy", "scifi"]

    def call(self) -> str:
        return f"{self.title} by {self.author}"


def test_gemini_tool() -> None:
    """Tests the `GeminiTool` class."""
    tool_call = FunctionCall(
        name="FormatBook",
        args={
            "title": "The Name of the Wind",
            "author": "Patrick Rothfuss",
            "genre": "fantasy",
        },
    )

    tool = FormatBook.from_tool_call(tool_call)
    assert isinstance(tool, BaseTool)
    assert isinstance(tool, GeminiTool)
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    assert tool.call() == "The Name of the Wind by Patrick Rothfuss"

    assert (
        FormatBook.tool_schema().to_proto()
        == Tool(
            function_declarations=[
                FunctionDeclaration(
                    name="FormatBook",
                    description="Returns the title and author nicely formatted.",
                    parameters={
                        "properties": {
                            "title": {"type": "string"},
                            "author": {"type": "string"},
                            "genre": {
                                "enum": ["fantasy", "scifi"],
                                "type": "string",
                                "format": "enum",
                            },
                        },
                        "required": ["title", "author", "genre"],
                        "type": "object",
                    },
                )
            ]
        ).to_proto()
    )


def test_gemini_tool_no_nesting() -> None:
    """Tests the `GeminiTool` class with nested structures."""

    class Def(Enum):
        A = "a"

    class Nested(GeminiTool):
        nested: Def

    with pytest.raises(
        ValueError,
        match="Unfortunately Google's Gemini API cannot handle nested structures "
        "with \\$defs.",
    ):
        Nested.tool_schema()



"""Tests the `vertex._utils.handle_stream` module."""

import pytest
from vertexai.generative_models import (
    Candidate,
    Content,
    Part,
)
from vertexai.generative_models import (
    GenerationResponse as GenerateContentResponse,
)

from mirascope.core.vertex._utils._handle_stream import (
    handle_stream,
    handle_stream_async,
)


@pytest.fixture()
def mock_chunks() -> list[GenerateContentResponse]:
    """Returns a list of mock `GenerateContentResponse` instances."""
    return [
        GenerateContentResponse.from_dict(
            {
                "candidates": [
                    Candidate.from_dict(
                        {
                            "finish_reason": 1,
                            "content": Content(
                                parts=[Part.from_text("The author is ")],
                                role="model",
                            ).to_dict(),
                        }
                    ).to_dict()
                ]
            }
        ),
        GenerateContentResponse.from_dict(
            {
                "candidates": [
                    Candidate.from_dict(
                        {
                            "finish_reason": 1,
                            "content": Content(
                                parts=[Part.from_text("Patrick Rothfuss")], role="model"
                            ).to_dict(),
                        }
                    ).to_dict()
                ]
            }
        ),
    ]


def test_handle_stream(mock_chunks: list[GenerateContentResponse]) -> None:
    """Tests the `handle_stream` function."""
    result = list(handle_stream((chunk for chunk in mock_chunks), tool_types=None))
    assert len(result) == 2
    assert result[0][0].content == "The author is "
    assert result[0][1] is None
    assert result[1][0].content == "Patrick Rothfuss"
    assert result[1][1] is None


@pytest.mark.asyncio
async def test_handle_stream_async(
    mock_chunks: list[GenerateContentResponse],
) -> None:
    """Tests the `handle_stream_async` function."""
    result = []

    async def generator():
        for chunk in mock_chunks:
            yield chunk

    async for chunk, tool in handle_stream_async(generator(), tool_types=None):
        result.append((chunk, tool))

    assert len(result) == 2
    assert result[0][0].content == "The author is "
    assert result[0][1] is None
    assert result[1][0].content == "Patrick Rothfuss"
    assert result[1][1] is None



"""Get JSON output from a Google response."""

import json

from proto.marshal.collections import RepeatedComposite

from ..call_response import GoogleCallResponse
from ..call_response_chunk import GoogleCallResponseChunk


def get_json_output(
    response: GoogleCallResponse | GoogleCallResponseChunk, json_mode: bool
) -> str:
    """Extracts the JSON output from a Google response."""
    if isinstance(response, GoogleCallResponse):
        if json_mode and (content := response.content):
            json_start = content.index("{")
            json_end = content.rfind("}")
            return content[json_start : json_end + 1]
        elif tool_calls := [
            function_call
            for function_call in (response.response.function_calls or [])  # pyright: ignore [reportOptionalSubscript, reportOptionalIterable, reportOptionalMemberAccess]
            if function_call.args  # pyright: ignore [reportOptionalSubscript, reportOptionalIterable, reportOptionalMemberAccess]
        ]:
            return json.dumps(
                {
                    k: v if not isinstance(v, RepeatedComposite) else list(v)
                    for k, v in (tool_calls[0].args or {}).items()  # pyright: ignore [reportOptionalMemberAccess]
                }
                if isinstance(tool_calls, list) and tool_calls[0]
                else {}
            )
        else:
            raise ValueError("No tool call or JSON object found in response.")
    elif not json_mode:
        raise ValueError("Google only supports structured streaming in json mode.")
    return response.content



"""This module contains the `XAICallResponse` class.

usage docs: learn/calls.md#handling-responses
"""

from ..openai import OpenAICallResponse


class XAICallResponse(OpenAICallResponse):
    """A simpler wrapper around `OpenAICallResponse`.

    Everything is the same except the `cost` property, which has been updated to use
    xAI's cost calculations so that cost tracking works for non-OpenAI models.
    """

    _provider = "xai"



"""The Mirascope Groq Module."""

from typing import TypeAlias

from groq.types.chat import ChatCompletionMessageParam

from ..base import BaseMessageParam
from ._call import groq_call
from ._call import groq_call as call
from .call_params import GroqCallParams
from .call_response import GroqCallResponse
from .call_response_chunk import GroqCallResponseChunk
from .dynamic_config import AsyncGroqDynamicConfig, GroqDynamicConfig
from .stream import GroqStream
from .tool import GroqTool

GroqMessageParam: TypeAlias = ChatCompletionMessageParam | BaseMessageParam

__all__ = [
    "AsyncGroqDynamicConfig",
    "GroqCallParams",
    "GroqCallResponse",
    "GroqCallResponseChunk",
    "GroqDynamicConfig",
    "GroqMessageParam",
    "GroqStream",
    "GroqTool",
    "call",
    "groq_call",
]



"""This module contains the OpenAIMessageParamConverter class."""

import json
from collections.abc import Iterable
from typing import cast

from openai.types.chat import ChatCompletionMessageParam

from mirascope.core import BaseMessageParam
from mirascope.core.base import (
    AudioPart,
    ImageURLPart,
    TextPart,
    ToolCallPart,
    ToolResultPart,
)
from mirascope.core.base._utils._base_message_param_converter import (
    BaseMessageParamConverter,
)
from mirascope.core.openai._utils import convert_message_params


class OpenAIMessageParamConverter(BaseMessageParamConverter):
    @staticmethod
    def to_provider(
        message_params: list[BaseMessageParam],
    ) -> list[ChatCompletionMessageParam]:
        """Converts base message params to OpenAI message params."""
        return convert_message_params(
            cast(list[BaseMessageParam | ChatCompletionMessageParam], message_params)
        )

    @staticmethod
    def from_provider(
        message_params: list[ChatCompletionMessageParam],
    ) -> list[BaseMessageParam]:
        """Converts OpenAI message params to base message params."""
        converted = []
        for message_param in message_params:
            contents = []
            content = message_param.get("content")
            if (
                message_param["role"] == "tool"
                and "name" in message_param
                and "content" in message_param
            ):
                converted.append(
                    BaseMessageParam(
                        role="tool",
                        content=[
                            ToolResultPart(
                                type="tool_result",
                                name=message_param["name"],
                                content=message_param["content"],
                                id=message_param["tool_call_id"],
                                is_error=False,
                            )
                        ],
                    )
                )
                continue
            elif isinstance(content, str):
                converted.append(
                    BaseMessageParam(role=message_param["role"], content=content)
                )
                continue
            elif isinstance(content, Iterable):
                for part in content:
                    if part["type"] == "text":
                        contents.append(TextPart(type="text", text=part["text"]))
                    elif part["type"] == "image_url":
                        image_url = part["image_url"]
                        contents.append(
                            ImageURLPart(
                                type="image_url",
                                url=image_url["url"],
                                detail=image_url.get("detail", None),
                            )
                        )
                    elif part["type"] == "input_audio":
                        input_audio = part["input_audio"]
                        contents.append(
                            AudioPart(
                                type="audio",
                                media_type=f"audio/{input_audio['format']}",
                                audio=input_audio["data"],
                            )
                        )
                    else:
                        raise ValueError(part["refusal"])  # pyright: ignore [reportGeneralTypeIssues]
            if tool_calls := message_param.get("tool_calls"):
                for tool_call in tool_calls:
                    contents.append(
                        ToolCallPart(
                            type="tool_call",
                            name=tool_call["function"]["name"],
                            id=tool_call["id"],
                            args=json.loads(tool_call["function"]["arguments"]),
                        )
                    )
            if contents:
                converted.append(
                    BaseMessageParam(role=message_param["role"], content=contents)
                )
        return converted



from __future__ import annotations

import inspect
from abc import ABC
from typing import ClassVar, Generic, TypeVar, cast

from pydantic import BaseModel, ConfigDict, create_model

from mirascope.core import BaseTool, BaseToolKit
from mirascope.core.base._utils import DEFAULT_TOOL_DOCSTRING


class _ConfigurableToolConfig(BaseModel, ABC):
    """Base configuration for tools"""

    @classmethod
    def from_env(cls: type[_ConfigurableToolConfigT]) -> _ConfigurableToolConfigT:
        """Returns a configuration instance from environment variables."""
        return cls()


_ConfigurableToolConfigT = TypeVar(
    "_ConfigurableToolConfigT", bound=_ConfigurableToolConfig
)

_ToolSchemaT = TypeVar("_ToolSchemaT")


class ConfigurableTool(BaseTool, Generic[_ConfigurableToolConfigT], ABC):
    """Abstract base class for configurable tools.

    Subclasses must define a `__prompt_usage_description__` class variable
    and __configurable_tool_config__ class variable with a subclass of _ToolConfig.
    """

    __configurable_tool_config__: ClassVar[_ConfigurableToolConfig]
    __prompt_usage_description__: ClassVar[str]
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def _get_config(cls) -> _ConfigurableToolConfigT:
        """Get tool configuration"""
        return cast(_ConfigurableToolConfigT, cls.__configurable_tool_config__)

    @classmethod
    def from_config(
        cls: type[_ConfigurableToolT], config: _ConfigurableToolConfigT
    ) -> type[_ConfigurableToolT]:
        """Create tool class with custom configuration"""

        # ClassVar cannot get TypeVar Type. So, we need the comment to ignore the error.
        config = cls._get_config().model_validate(config)  # pyright: ignore [reportAssignmentType]
        new_model = create_model(
            cls.__name__,
            __base__=cls,
            __module__=cls.__module__,
            __doc__=cls.__doc__ if cls.__doc__ else DEFAULT_TOOL_DOCSTRING,
        )
        new_model.__prompt_usage_description__ = cls.__prompt_usage_description__
        new_model.__configurable_tool_config__ = config
        return new_model

    @classmethod
    def usage_description(cls) -> str:
        """Returns instructions for using this tool."""
        return inspect.cleandoc(cls.__prompt_usage_description__)


_ConfigurableToolT = TypeVar("_ConfigurableToolT", bound=ConfigurableTool)


class ConfigurableToolKit(BaseToolKit, Generic[_ConfigurableToolConfigT], ABC):
    """Abstract base class for configurable toolkit.

    Subclasses must define a `__prompt_usage_description__` class variable
    and __config__ class variable with a subclass of _ToolConfig.
    """

    config: _ConfigurableToolConfigT
    __prompt_usage_description__: ClassVar[str]
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def create_tools(self) -> list[type[BaseTool]]:
        """The method to create the tools."""
        return [
            tool.from_config(self.config)
            if issubclass(tool, ConfigurableTool) and "config" in self.model_fields_set
            else tool
            for tool in super().create_tools()
        ]

    @classmethod
    def usage_description(cls) -> str:
        """Returns instructions for using this toolkit."""
        return inspect.cleandoc(cls.__prompt_usage_description__)


_ConfigurableToolKitT = TypeVar("_ConfigurableToolKitT", bound=ConfigurableToolKit)



from typing import ClassVar

from pydantic import Field

from mirascope.tools.base import ConfigurableTool, _ConfigurableToolConfig


class MockConfigConfigurable(_ConfigurableToolConfig):
    value: str = Field("default")


class MockTool(ConfigurableTool[MockConfigConfigurable]):
    __configurable_tool_config__ = MockConfigConfigurable()  # pyright: ignore [reportCallIssue]
    __prompt_usage_description__: ClassVar[str] = "Test description"

    input: str = Field(..., description="Test input")

    def call(self) -> str:
        return f"Test output: {self.input}"


def test_tool_config():
    config = MockConfigConfigurable()  # pyright: ignore [reportCallIssue]
    assert config.value == "default"

    custom_config = MockConfigConfigurable(value="custom")
    assert custom_config.value == "custom"

    env_config = MockConfigConfigurable.from_env()
    assert isinstance(env_config, MockConfigConfigurable)


def test_configurable_tool():
    tool = MockTool(input="test")
    result = tool.call()
    assert result == "Test output: test"


def test_tool_config_access():
    tool = MockTool(input="test")
    config = tool._get_config()
    assert isinstance(config, MockConfigConfigurable)
    assert config.value == "default"


def test_configurable_tool_from_config():
    custom_config = MockConfigConfigurable(value="custom")
    CustomTool = MockTool.from_config(custom_config)

    tool = CustomTool(input="test")  # pyright: ignore [reportCallIssue]
    assert isinstance(tool._get_config(), MockConfigConfigurable)
    assert tool._get_config().value == "custom"

    result = tool.call()
    assert result == "Test output: test"



"""This module contains the `MistralCallResponse` class.

usage docs: learn/calls.md#handling-responses
"""

from collections.abc import Sequence
from functools import cached_property
from typing import Any, cast

from mistralai import ChatCompletionChoice
from mistralai.models import (
    AssistantMessage,
    ChatCompletionResponse,
    SystemMessage,
    ToolMessage,
    UsageInfo,
    UserMessage,
)
from pydantic import computed_field

from .. import BaseMessageParam
from ..base import BaseCallResponse, transform_tool_outputs
from ..base.types import CostMetadata, FinishReason
from ._utils._convert_finish_reason_to_common_finish_reasons import (
    _convert_finish_reasons_to_common_finish_reasons,
)
from ._utils._message_param_converter import MistralMessageParamConverter
from .call_params import MistralCallParams
from .dynamic_config import MistralDynamicConfig
from .tool import MistralTool


class MistralCallResponse(
    BaseCallResponse[
        ChatCompletionResponse,
        MistralTool,
        dict[str, Any],
        MistralDynamicConfig,
        AssistantMessage | SystemMessage | ToolMessage | UserMessage,
        MistralCallParams,
        UserMessage,
        MistralMessageParamConverter,
    ]
):
    """A convenience wrapper around the Mistral `ChatCompletion` response.

    When calling the Mistral API using a function decorated with `mistral_call`, the
    response will be an `MistralCallResponse` instance with properties that allow for
    more convenience access to commonly used attributes.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.mistral import mistral_call


    @mistral_call("mistral-largel-latest")
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"

    response = recommend_book("fantasy")  # response is an `MistralCallResponse` instance
    print(response.content)
    ```
    """

    _message_converter: type[MistralMessageParamConverter] = (
        MistralMessageParamConverter
    )

    _provider = "mistral"

    @property
    def _response_choices(self) -> list[ChatCompletionChoice]:
        return self.response.choices or []

    @computed_field
    @property
    def content(self) -> str:
        """The content of the chat completion for the 0th choice."""
        return cast(str, self._response_choices[0].message.content) or ""

    @computed_field
    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        return [
            choice.finish_reason if choice.finish_reason else ""
            for choice in self._response_choices
        ]

    @computed_field
    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return self.response.model

    @computed_field
    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.response.id

    @property
    def usage(self) -> UsageInfo:
        """Returns the usage of the chat completion."""
        return self.response.usage

    @computed_field
    @property
    def input_tokens(self) -> int:
        """Returns the number of input tokens."""
        return self.usage.prompt_tokens

    @computed_field
    @property
    def cached_tokens(self) -> int:
        """Returns the number of cached tokens."""
        return 0

    @computed_field
    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        return self.usage.completion_tokens

    @computed_field
    @cached_property
    def message_param(self) -> AssistantMessage:
        """Returns the assistants's response as a message parameter."""
        return self._response_choices[0].message

    @cached_property
    def tools(self) -> list[MistralTool] | None:
        """Returns the tools for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        tool_calls = self._response_choices[0].message.tool_calls
        if not self.tool_types or not tool_calls:
            return None

        extracted_tools = []
        for tool_call in tool_calls:
            for tool_type in self.tool_types:
                if tool_call.function.name == tool_type._name():
                    extracted_tools.append(tool_type.from_tool_call(tool_call))
                    break

        return extracted_tools

    @cached_property
    def tool(self) -> MistralTool | None:
        """Returns the 0th tool for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        if tools := self.tools:
            return tools[0]
        return None

    @classmethod
    @transform_tool_outputs
    def tool_message_params(
        cls, tools_and_outputs: Sequence[tuple[MistralTool, str]]
    ) -> list[ToolMessage]:
        """Returns the tool message parameters for tool call results.

        Args:
            tools_and_outputs: The sequence of tools and their outputs from which the tool
                message parameters should be constructed.

        Returns:
            The list of constructed `ChatMessage` parameters.
        """
        return [
            ToolMessage(
                content=output,
                tool_call_id=tool.tool_call.id,  # pyright: ignore [reportOptionalMemberAccess]
                name=tool._name(),
            )
            for tool, output in tools_and_outputs
        ]

    @property
    def common_finish_reasons(self) -> list[FinishReason] | None:
        return _convert_finish_reasons_to_common_finish_reasons(self.finish_reasons)

    @property
    def common_message_param(self) -> BaseMessageParam:
        return MistralMessageParamConverter.from_provider([self.message_param])[0]

    @property
    def common_user_message_param(self) -> BaseMessageParam | None:
        if not self.user_message_param:
            return None
        return MistralMessageParamConverter.from_provider([self.user_message_param])[0]

    @computed_field
    @property
    def cost_metadata(self) -> CostMetadata:
        """Get metadata required for cost calculation."""
        return super().cost_metadata



from collections.abc import Sequence
from functools import cached_property
from typing import Any
from unittest.mock import Mock, patch

import pytest
from openai import AsyncOpenAI, OpenAI
from pydantic import computed_field

from mirascope.core.base import (
    BaseCallKwargs,
    BaseCallParams,
    BaseCallResponse,
    BaseStream,
    BaseTool,
    Metadata,
)
from mirascope.core.base._utils import BaseMessageParamConverter
from mirascope.core.base.types import CostMetadata, FinishReason
from mirascope.llm._call import (
    _get_local_provider_call,
    _get_provider_call,
    _wrap_result,
    call,
)
from mirascope.llm._context import context
from mirascope.llm.call_response import CallResponse
from mirascope.llm.stream import Stream


class DummyCallParams(BaseCallParams): ...


class ConcreteResponse(BaseCallResponse[Any, Any, Any, Any, Any, Any, Any, Any]):
    _message_converter: type = BaseMessageParamConverter

    @property
    def content(self): ...  # pyright: ignore [reportIncompatibleMethodOverride]

    @property
    def finish_reasons(self): ...

    @property
    def model(self): ...

    @property
    def id(self): ...

    @property
    def usage(self): ...

    @property
    def input_tokens(self): ...

    @property
    def cached_tokens(self): ...

    @property
    def output_tokens(self): ...

    @property
    def cost(self): ...

    @computed_field
    @cached_property
    def message_param(self) -> Any: ...

    @cached_property
    def tools(self) -> Any: ...

    @cached_property
    def tool(self) -> Any: ...

    @classmethod
    def tool_message_params(cls, tools_and_outputs: Sequence[tuple[BaseTool, str]]): ...  # pyright: ignore [reportIncompatibleMethodOverride]

    @property
    def common_finish_reasons(self) -> list[FinishReason] | None:
        # Just return ["stop"] as a string to avoid AttributeError
        return ["stop"]

    @property
    def common_message_param(self): ...  # pyright: ignore [reportIncompatibleMethodOverride]

    @property
    def common_user_message_param(self): ...

    @property
    def common_tools(self): ...

    @property
    def common_usage(self): ...

    def common_construct_call_response(self): ...

    def common_construct_message_param(
        self, tool_calls: list | None, content: str | None
    ): ...

    @property
    def cost_metadata(self) -> CostMetadata: ...


class ConcreteStream(BaseStream):
    def _construct_message_param(self, *args, **kwargs): ...

    def construct_call_response(self): ...

    @property
    def cost(self): ...

    @property
    def cost_metadata(self) -> CostMetadata: ...


def test_wrap_result():
    resp = ConcreteResponse(
        metadata=Metadata(),
        response={},
        tool_types=None,
        prompt_template=None,
        fn_args={},
        dynamic_config={},
        messages=[],
        call_params=DummyCallParams(),
        call_kwargs=BaseCallKwargs(),
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    result = _wrap_result(resp)
    assert isinstance(result, CallResponse)
    assert result.finish_reasons == ["stop"]

    strm = ConcreteStream(
        stream=Mock(),
        metadata=Metadata(),
        tool_types=Mock(),
        call_response_type=Mock(),  # pyright: ignore [reportArgumentType]
        model=Mock(),  # pyright: ignore [reportArgumentType]
        prompt_template=Mock(),
        fn_args={},
        dynamic_config={},
        messages=[],
        call_params=DummyCallParams(),
        call_kwargs=BaseCallKwargs(),
    )
    result_stream = _wrap_result(strm)
    assert isinstance(result_stream, Stream)

    output = _wrap_result("parsed output")
    assert isinstance(output, str)
    assert output == "parsed output"


def test_wrap_result_with_base_tool():
    """Test that _wrap_result can handle a response containing a BaseTool."""

    # Create a custom BaseTool for testing
    class CustomBaseTool(BaseTool): ...

    # Create a response instance
    resp = ConcreteResponse(
        metadata=Metadata(),
        response={},
        tool_types=[CustomBaseTool],  # Use the BaseTool subclass
        prompt_template=None,
        fn_args={},
        dynamic_config={},
        messages=[],
        call_params=DummyCallParams(),
        call_kwargs=BaseCallKwargs(),
        user_message_param=None,
        start_time=0,
        end_time=0,
    )

    result = _wrap_result(resp)
    assert isinstance(result, CallResponse)

    # Create a stream instance
    resp = ConcreteStream(
        stream=Mock(),
        metadata=Metadata(),
        tool_types=[CustomBaseTool],  # Use the BaseTool subclass
        call_response_type=Mock(),  # pyright: ignore [reportArgumentType]
        model=Mock(),  # pyright: ignore [reportArgumentType]
        prompt_template=Mock(),
        fn_args={},
        dynamic_config={},
        messages=[],
        call_params=DummyCallParams(),
        call_kwargs=BaseCallKwargs(),
    )

    result = _wrap_result(resp)
    assert isinstance(result, Stream)


def test_get_provider_call_unsupported():
    with pytest.raises(ValueError, match="Unsupported provider: bad_provider"):
        _get_provider_call("bad_provider")  # pyright: ignore [reportArgumentType]


# Use `new=` instead of `return_value=` so that the call function is replaced by a string directly.
def test_get_provider_call_anthropic():
    with patch("mirascope.core.anthropic.anthropic_call", new="anthropic_mock"):
        func = _get_provider_call("anthropic")
        assert func == "anthropic_mock"


def test_get_provider_call_azure():
    with patch("mirascope.core.azure.azure_call", new="azure_mock"):
        func = _get_provider_call("azure")
        assert func == "azure_mock"


def test_get_provider_call_bedrock():
    with patch("mirascope.core.bedrock.bedrock_call", new="bedrock_mock"):
        func = _get_provider_call("bedrock")
        assert func == "bedrock_mock"


def test_get_provider_call_cohere():
    with patch("mirascope.core.cohere.cohere_call", new="cohere_mock"):
        func = _get_provider_call("cohere")
        assert func == "cohere_mock"


def test_get_provider_call_gemini():
    with patch("mirascope.core.gemini.gemini_call", new="gemini_mock"):
        func = _get_provider_call("gemini")
        assert func == "gemini_mock"


def test_get_provider_call_google():
    with patch("mirascope.core.google.google_call", new="google_mock"):
        func = _get_provider_call("google")
        assert func == "google_mock"


def test_get_provider_call_groq():
    with patch("mirascope.core.groq.groq_call", new="groq_mock"):
        func = _get_provider_call("groq")
        assert func == "groq_mock"


def test_get_provider_call_litellm():
    with patch("mirascope.core.litellm.litellm_call", new="litellm_mock"):
        func = _get_provider_call("litellm")
        assert func == "litellm_mock"


def test_get_provider_call_mistral():
    with patch("mirascope.core.mistral.mistral_call", new="mistral_mock"):
        func = _get_provider_call("mistral")
        assert func == "mistral_mock"


def test_get_provider_call_openai():
    with patch("mirascope.core.openai.openai_call", new="openai_mock"):
        func = _get_provider_call("openai")
        assert func == "openai_mock"


def test_get_provider_call_vertex():
    with patch("mirascope.core.vertex.vertex_call", new="vertex_mock"):
        func = _get_provider_call("vertex")
        assert func == "vertex_mock"


def test_get_provider_call_xai():
    with patch("mirascope.core.xai.xai_call", new="xai_mock"):
        func = _get_provider_call("xai")
        assert func == "xai_mock"


def test_get_local_provider_call_ollama():
    with patch("mirascope.core.openai.openai_call", new="openai_ollama_mock"):
        func, client = _get_local_provider_call("ollama", None, False)
        assert func == "openai_ollama_mock"
        assert (
            isinstance(client, OpenAI)
            and client.api_key == "ollama"
            and str(client.base_url) == "http://localhost:11434/v1/"
        )
        func, client = _get_local_provider_call("ollama", None, True)
        assert func == "openai_ollama_mock"
        assert (
            isinstance(client, AsyncOpenAI)
            and client.api_key == "ollama"
            and str(client.base_url) == "http://localhost:11434/v1/"
        )
        mock_client = Mock()
        _, client = _get_local_provider_call("ollama", mock_client, False)
        assert client == mock_client


def test_get_local_provider_call_vllm():
    with patch("mirascope.core.openai.openai_call", new="openai_vllm_mock"):
        func, client = _get_local_provider_call("vllm", None, False)
        assert func == "openai_vllm_mock"
        assert (
            isinstance(client, OpenAI)
            and client.api_key == "ollama"
            and str(client.base_url) == "http://localhost:8000/v1/"
        )
        func, client = _get_local_provider_call("vllm", None, True)
        assert func == "openai_vllm_mock"
        assert (
            isinstance(client, AsyncOpenAI)
            and client.api_key == "ollama"
            and str(client.base_url) == "http://localhost:8000/v1/"
        )
        mock_client = Mock()
        _, client = _get_local_provider_call("vllm", mock_client, False)
        assert client == mock_client


def test_call_decorator_sync():
    def dummy_provider_call(
        model,
        stream,
        tools,
        response_model,
        output_parser,
        json_mode,
        call_params,
        client,
    ):
        def wrapper(fn):
            def inner(*args, **kwargs):
                return ConcreteResponse(
                    metadata=Metadata(),
                    response={},
                    tool_types=None,
                    prompt_template=None,
                    fn_args={},
                    dynamic_config={},
                    messages=[],
                    call_params=DummyCallParams(),
                    call_kwargs=BaseCallKwargs(),
                    user_message_param=None,
                    start_time=0,
                    end_time=0,
                )

            return inner

        return wrapper

    with patch(
        "mirascope.llm._call._get_provider_call", return_value=dummy_provider_call
    ):

        @call(provider="openai", model="gpt-4o-mini")
        def dummy_function(): ...

        res = dummy_function()
        assert isinstance(res, CallResponse)
        # finish_reasons is ["stop"] due to our override
        assert res.finish_reasons == ["stop"]

    with patch(
        "mirascope.llm._call._get_local_provider_call",
        return_value=(dummy_provider_call, None),
    ):

        @call(provider="ollama", model="gpt-4o-mini")
        def dummy_local_function(): ...

        res = dummy_local_function()
        assert isinstance(res, CallResponse)
        assert res.finish_reasons == ["stop"]


@pytest.mark.asyncio
async def test_call_decorator_async():
    def dummy_async_provider_call(
        model,
        stream,
        tools,
        response_model,
        output_parser,
        json_mode,
        call_params,
        client,
    ):
        def wrapper(fn):
            async def inner(*args, **kwargs):
                return ConcreteResponse(
                    metadata=Metadata(),
                    response={},
                    tool_types=None,
                    prompt_template=None,
                    fn_args={},
                    dynamic_config={},
                    messages=[],
                    call_params=DummyCallParams(),
                    call_kwargs=BaseCallKwargs(),
                    user_message_param=None,
                    start_time=0,
                    end_time=0,
                )

            return inner

        return wrapper

    with patch(
        "mirascope.llm._call._get_provider_call",
        return_value=dummy_async_provider_call,
    ):

        @call(provider="openai", model="gpt-4o-mini")
        async def dummy_async_function(): ...

        res = await dummy_async_function()
        assert isinstance(res, CallResponse)
        assert res.finish_reasons == ["stop"]

    with patch(
        "mirascope.llm._call._get_local_provider_call",
        return_value=(dummy_async_provider_call, None),
    ):

        @call(provider="ollama", model="gpt-4o-mini")
        async def dummy_local_async_function(): ...

        res = await dummy_local_async_function()
        assert isinstance(res, CallResponse)
        assert res.finish_reasons == ["stop"]


@pytest.mark.asyncio
async def test_context_in_async_function():
    """Test that context is properly applied in async functions."""
    # Create a mock provider call that captures the effective call args
    captured_args = {}

    def dummy_async_provider_call(
        model,
        stream,
        tools,
        response_model,
        output_parser,
        json_mode,
        call_params,
        client,
    ):
        def wrapper(fn):
            async def inner(*args, **kwargs):
                # Store the args that were passed to the provider call
                nonlocal captured_args
                captured_args = {
                    "model": model,
                    "stream": stream,
                    "tools": tools,
                    "response_model": response_model,
                    "output_parser": output_parser,
                    "json_mode": json_mode,
                    "call_params": call_params,
                    "client": client,
                }

                return ConcreteResponse(
                    metadata=Metadata(),
                    response={},
                    tool_types=None,
                    prompt_template=None,
                    fn_args={},
                    dynamic_config={},
                    messages=[],
                    call_params=DummyCallParams(),
                    call_kwargs=BaseCallKwargs(),
                    user_message_param=None,
                    start_time=0,
                    end_time=0,
                )

            return inner

        return wrapper

    with patch(
        "mirascope.llm._call._get_provider_call",
        return_value=dummy_async_provider_call,
    ):
        # Create a function with the openai provider
        @call(provider="openai", model="gpt-4o-mini")
        async def dummy_async_function():
            pass  # pragma: no cover

        # Call the function with a context that overrides the model
        with context(provider="openai", model="gpt-4o"):
            await dummy_async_function()

        # Check that the context override was applied
        assert captured_args["model"] == "gpt-4o", (
            "Context model override was not applied in async function"
        )


@pytest.mark.asyncio
async def test_context_in_async_function_with_gather():
    """Test that context is properly applied in async functions when using asyncio.gather."""
    # Create a mock provider call that captures the effective call args for each call
    captured_args_list = []

    def dummy_async_provider_call(
        model,
        stream,
        tools,
        response_model,
        output_parser,
        json_mode,
        call_params,
        client,
    ):
        def wrapper(fn):
            async def inner(*args, **kwargs):
                # Store the args that were passed to the provider call
                captured_args = {
                    "model": model,
                    "stream": stream,
                    "tools": tools,
                    "response_model": response_model,
                    "output_parser": output_parser,
                    "json_mode": json_mode,
                    "call_params": call_params,
                    "client": client,
                }
                captured_args_list.append(captured_args)

                return ConcreteResponse(
                    metadata=Metadata(),
                    response={},
                    tool_types=None,
                    prompt_template=None,
                    fn_args={},
                    dynamic_config={},
                    messages=[],
                    call_params=DummyCallParams(),
                    call_kwargs=BaseCallKwargs(),
                    user_message_param=None,
                    start_time=0,
                    end_time=0,
                )

            return inner

        return wrapper

    with patch(
        "mirascope.llm._call._get_provider_call",
        return_value=dummy_async_provider_call,
    ):
        # Create a function with the openai provider
        @call(provider="openai", model="gpt-4o-mini")
        async def dummy_async_function():
            pass  # pragma: no cover

        # Create futures first, then await them together
        import asyncio

        # Create the first future with default provider/model
        future1 = dummy_async_function()

        # Create the second future with a different context
        with context(provider="anthropic", model="claude-3-5-sonnet"):
            future2 = dummy_async_function()

        # Await both futures together
        await asyncio.gather(future1, future2)

        # Check that we have two captured args
        assert len(captured_args_list) == 2

        # The first should use the original model
        assert captured_args_list[0]["model"] == "gpt-4o-mini"

        # The second should use the context-overridden model
        assert captured_args_list[1]["model"] == "claude-3-5-sonnet", (
            "Context model override was not applied when using asyncio.gather"
        )



"""Calculate the cost of a Gemini API call."""

from ..base.types import CostMetadata

# Standard Gemini API pricing table
GEMINI_API_PRICING: dict[str, dict[str, float]] = {
    "gemini-2.0-pro": {
        "prompt_short": 0.000_001_25,
        "completion_short": 0.000_005,
        "prompt_long": 0.000_002_5,
        "completion_long": 0.000_01,
        "cached": 0.000_000_625,
    },
    "gemini-2.0-pro-preview-1206": {
        "prompt_short": 0.000_001_25,
        "completion_short": 0.000_005,
        "prompt_long": 0.000_002_5,
        "completion_long": 0.000_01,
        "cached": 0.000_000_625,
    },
    "gemini-2.0-flash": {
        "prompt_short": 0.000_000_10,
        "completion_short": 0.000_000_40,
        "prompt_long": 0.000_000_10,
        "completion_long": 0.000_000_40,
        "cached": 0.000_000_037_5,
    },
    "gemini-2.0-flash-latest": {
        "prompt_short": 0.000_000_10,
        "completion_short": 0.000_000_40,
        "prompt_long": 0.000_000_10,
        "completion_long": 0.000_000_40,
        "cached": 0.000_000_037_5,
    },
    "gemini-2.0-flash-001": {
        "prompt_short": 0.000_000_10,
        "completion_short": 0.000_000_40,
        "prompt_long": 0.000_000_10,
        "completion_long": 0.000_000_40,
        "cached": 0.000_000_037_5,
    },
    "gemini-2.0-flash-lite": {
        "prompt_short": 0.000_000_075,
        "completion_short": 0.000_000_30,
        "prompt_long": 0.000_000_075,
        "completion_long": 0.000_000_30,
        "cached": 0.000_000_037_5,
    },
    "gemini-2.0-flash-lite-preview-02-05": {
        "prompt_short": 0.000_000_075,
        "completion_short": 0.000_000_30,
        "prompt_long": 0.000_000_075,
        "completion_long": 0.000_000_30,
        "cached": 0.000_000_037_5,
    },
    "gemini-1.5-pro": {
        "prompt_short": 0.000_001_25,
        "completion_short": 0.000_005,
        "prompt_long": 0.000_002_5,
        "completion_long": 0.000_01,
        "cached": 0.000_000_625,
    },
    "gemini-1.5-pro-latest": {
        "prompt_short": 0.000_001_25,
        "completion_short": 0.000_005,
        "prompt_long": 0.000_002_5,
        "completion_long": 0.000_01,
        "cached": 0.000_000_625,
    },
    "gemini-1.5-pro-001": {
        "prompt_short": 0.000_001_25,
        "completion_short": 0.000_005,
        "prompt_long": 0.000_002_5,
        "completion_long": 0.000_01,
        "cached": 0.000_000_625,
    },
    "gemini-1.5-pro-002": {
        "prompt_short": 0.000_001_25,
        "completion_short": 0.000_005,
        "prompt_long": 0.000_002_5,
        "completion_long": 0.000_01,
        "cached": 0.000_000_625,
    },
    "gemini-1.5-flash": {
        "prompt_short": 0.000_000_075,
        "completion_short": 0.000_000_30,
        "prompt_long": 0.000_000_15,
        "completion_long": 0.000_000_60,
        "cached": 0.000_000_037_5,
    },
    "gemini-1.5-flash-latest": {
        "prompt_short": 0.000_000_075,
        "completion_short": 0.000_000_30,
        "prompt_long": 0.000_000_15,
        "completion_long": 0.000_000_60,
        "cached": 0.000_000_037_5,
    },
    "gemini-1.5-flash-001": {
        "prompt_short": 0.000_000_075,
        "completion_short": 0.000_000_30,
        "prompt_long": 0.000_000_15,
        "completion_long": 0.000_000_60,
        "cached": 0.000_000_037_5,
    },
    "gemini-1.5-flash-002": {
        "prompt_short": 0.000_000_075,
        "completion_short": 0.000_000_30,
        "prompt_long": 0.000_000_15,
        "completion_long": 0.000_000_60,
        "cached": 0.000_000_037_5,
    },
    "gemini-1.5-flash-8b": {
        "prompt_short": 0.000_000_037_5,
        "completion_short": 0.000_000_15,
        "prompt_long": 0.000_000_075,
        "completion_long": 0.000_000_30,
        "cached": 0.000_000_025,
    },
    "gemini-1.5-flash-8b-latest": {
        "prompt_short": 0.000_000_037_5,
        "completion_short": 0.000_000_15,
        "prompt_long": 0.000_000_075,
        "completion_long": 0.000_000_30,
        "cached": 0.000_000_025,
    },
    "gemini-1.5-flash-8b-001": {
        "prompt_short": 0.000_000_037_5,
        "completion_short": 0.000_000_15,
        "prompt_long": 0.000_000_075,
        "completion_long": 0.000_000_30,
        "cached": 0.000_000_025,
    },
    "gemini-1.5-flash-8b-002": {
        "prompt_short": 0.000_000_037_5,
        "completion_short": 0.000_000_15,
        "prompt_long": 0.000_000_075,
        "completion_long": 0.000_000_30,
        "cached": 0.000_000_025,
    },
    "gemini-1.0-pro": {
        "prompt_short": 0.000_000_5,
        "completion_short": 0.000_001_5,
        "prompt_long": 0.000_000_5,
        "completion_long": 0.000_001_5,
        "cached": 0.000_000,
    },
}

# Vertex AI pricing table
VERTEX_AI_PRICING: dict[str, dict[str, float]] = {
    "gemini-2.0-flash": {
        "text_input": 0.000_000_15,
        "image_input": 0.000_000_15,
        "video_input": 0.000_000_15,
        "audio_input": 0.000_001_00,
        "output": 0.000_000_60,
        "cached": 0.000_000_037_5,
        "cache_storage_per_hour": 0.000_001_00,
    },
    "gemini-2.0-flash-lite": {
        "text_input": 0.000_000_075,
        "image_input": 0.000_000_075,
        "video_input": 0.000_000_075,
        "audio_input": 0.000_000_075,
        "output": 0.000_000_30,
        "cached": 0.000_000_037_5,
        "cache_storage_per_hour": 0.000_001_00,
    },
    # Vertex AI pricing for Gemini 1.5 models is based on modalities rather than tokens
    "gemini-1.5-flash": {
        "text_input": 0.000_000_075,  # per 1K chars (approx. 250 tokens)
        "image_input": 0.000_02,  # per image
        "video_input": 0.000_02,  # per second
        "audio_input": 0.000_002,  # per second
        "output": 0.000_000_30,  # per 1K chars
        "cached_text": 0.000_000_046_875,  # per 1K chars
        "cached_image": 0.000_005,  # per image
        "cached_video": 0.000_005,  # per second
        "cached_audio": 0.000_000_5,  # per second
        "cache_storage_text": 0.000_25,  # per 1K chars per hour
        "cache_storage_image": 0.000_263,  # per image per hour
        "cache_storage_video": 0.000_263,  # per second per hour
        "cache_storage_audio": 0.000_025,  # per second per hour
    },
    "gemini-1.5-pro": {
        "text_input": 0.000_001_25,  # per 1K chars (approx. 250 tokens)
        "image_input": 0.000_32875,  # per image
        "video_input": 0.000_32875,  # per second
        "audio_input": 0.000_03125,  # per second
        "output": 0.000_005,  # per 1K chars
        "cached_text": 0.000_000_078125,  # per 1K chars
        "cached_image": 0.000_0821875,  # per image
        "cached_video": 0.000_0821875,  # per second
        "cached_audio": 0.000_0078125,  # per second
        "cache_storage_text": 0.001125,  # per 1K chars per hour
        "cache_storage_image": 0.0011835,  # per image per hour
        "cache_storage_video": 0.0011835,  # per second per hour
        "cache_storage_audio": 0.0001125,  # per second per hour
    },
}


def _calculate_context_cache_cost(
    metadata: CostMetadata,
    model_pricing: dict[str, float],
    model: str,
    use_vertex_ai: bool = False,
) -> float:
    """Calculate cost for context caching."""
    if metadata.context_cache_tokens is None or metadata.context_cache_hours is None:
        return 0.0

    if use_vertex_ai:
        # Vertex AI pricing depends on the model family
        if model.startswith("gemini-2.0"):
            return (
                metadata.context_cache_tokens
                * model_pricing.get("cache_storage_per_hour", 0)
                * metadata.context_cache_hours
            )
        elif model.startswith("gemini-1.5"):
            # Convert cache tokens to characters (approx)
            cache_chars = metadata.context_cache_tokens * 4
            return (
                (cache_chars / 1000)
                * model_pricing["cache_storage_text"]
                * metadata.context_cache_hours
            )

    # Standard Gemini API pricing - storage cost per token-hour
    storage_rate_per_token = 0.000001  # $1.00 per million tokens per hour
    if "flash-8b" in model:
        storage_rate_per_token = 0.00000025  # $0.25 per million tokens for 8B models

    return (
        metadata.context_cache_tokens
        * storage_rate_per_token
        * metadata.context_cache_hours
    )


def _calculate_grounding_cost(metadata: CostMetadata, model: str) -> float:
    """Calculate cost for grounding requests."""
    if metadata.google is None or metadata.google.grounding_requests is None:
        return 0.0

    # First 1,500 requests per day are free for Gemini 2.0 Flash models in Vertex AI
    if (
        model == "gemini-2.0-flash"
        and metadata.google.use_vertex_ai
        and metadata.google.grounding_requests <= 1500
    ):
        return 0.0

    # $35 per 1,000 requests for excess
    if metadata.google.use_vertex_ai and model == "gemini-2.0-flash":
        excess_requests = max(0, metadata.google.grounding_requests - 1500)
    else:
        excess_requests = metadata.google.grounding_requests

    return (excess_requests / 1000) * 35.0


def _calculate_vertex_2_0_cost(
    metadata: CostMetadata, model_pricing: dict[str, float], model: str
) -> float:
    """Calculate cost for Vertex AI's Gemini 2.0 models."""
    # Text tokens cost
    prompt_cost = (metadata.input_tokens or 0) * model_pricing["text_input"]
    completion_cost = (metadata.output_tokens or 0) * model_pricing["output"]
    cached_cost = (metadata.cached_tokens or 0) * model_pricing.get("cached", 0)

    # Context cache costs
    context_cache_cost = _calculate_context_cache_cost(
        metadata, model_pricing, model, use_vertex_ai=True
    )

    # Grounding costs
    grounding_cost = _calculate_grounding_cost(metadata, model)

    # Apply batch mode discount (50% for Vertex AI)
    if metadata.batch_mode:
        prompt_cost *= 0.5
        completion_cost *= 0.5
        context_cache_cost *= 0.5
        # Note: We don't discount grounding costs

    total_cost = (
        prompt_cost
        + completion_cost
        + cached_cost
        + context_cache_cost
        + grounding_cost
    )

    return total_cost


def _calculate_vertex_1_5_cost(
    metadata: CostMetadata, model_pricing: dict[str, float], model: str
) -> float:
    """Calculate cost for Vertex AI's Gemini 1.5 models."""
    # Text cost - convert tokens to characters (approx 4 chars per token)
    text_chars = (metadata.input_tokens or 0) * 4  # Approximation
    text_cost = (text_chars / 1000) * model_pricing["text_input"]

    # Output cost
    output_chars = (metadata.output_tokens or 0) * 4  # Approximation
    output_cost = (output_chars / 1000) * model_pricing["output"]

    # Context cache costs
    context_cache_cost = _calculate_context_cache_cost(
        metadata, model_pricing, model, use_vertex_ai=True
    )

    # Grounding costs
    grounding_cost = _calculate_grounding_cost(metadata, model)

    # Apply batch mode discount if applicable (50% off for Vertex AI)
    if metadata.batch_mode:
        text_cost *= 0.5
        output_cost *= 0.5
        context_cache_cost *= 0.5
        # Note: We don't discount grounding costs

    total_cost = text_cost + output_cost + context_cache_cost + grounding_cost

    return total_cost


def _calculate_standard_gemini_cost(
    metadata: CostMetadata,
    model_pricing: dict[str, float],
    model: str,
    use_long_context: bool,
) -> float:
    """Calculate cost for standard Gemini API."""
    # Determine prices based on context length
    prompt_price = model_pricing["prompt_long" if use_long_context else "prompt_short"]
    cached_price = model_pricing["cached"]
    completion_price = model_pricing[
        "completion_long" if use_long_context else "completion_short"
    ]

    # Basic token costs
    prompt_cost = (metadata.input_tokens or 0) * prompt_price
    cached_cost = (metadata.cached_tokens or 0) * cached_price
    completion_cost = (metadata.output_tokens or 0) * completion_price

    # Media token costs is included in the prompt/completion cost

    # Context cache costs
    context_cache_cost = _calculate_context_cache_cost(
        metadata, model_pricing, model, use_vertex_ai=False
    )

    # Grounding costs - only applies to certain models
    grounding_cost = _calculate_grounding_cost(metadata, model)

    total_cost = (
        prompt_cost
        + cached_cost
        + completion_cost
        + context_cache_cost
        + grounding_cost
    )

    return total_cost


def calculate_cost(
    metadata: CostMetadata,
    model: str,
) -> float | None:
    """Calculate the cost of a Google API call.

    This function supports both direct Gemini API and Vertex AI pricing.
    It handles different media types (text, image, video, audio) and special features
    like context caching and grounding.

    https://ai.google.dev/pricing
    https://cloud.google.com/vertex-ai/generative-ai/pricing

    Args:
        metadata: Additional metadata required for cost calculation
        model: Model name to use for pricing calculation

    Returns:
        Total cost in USD or None if invalid input
    """
    # Basic validation
    if metadata.input_tokens is None or metadata.output_tokens is None:
        return None

    # Initialize default values
    if metadata.cached_tokens is None:
        metadata.cached_tokens = 0

    # Check if we're using Vertex AI pricing
    use_vertex_ai = metadata.google and metadata.google.use_vertex_ai

    # Determine if we're using long context pricing
    use_long_context = (
        metadata.context_length is not None and metadata.context_length > 128_000
    ) or (metadata.input_tokens > 128_000)

    # Get the appropriate pricing table
    try:
        if use_vertex_ai and model in VERTEX_AI_PRICING:
            model_pricing = VERTEX_AI_PRICING[model]
        else:
            model_pricing = GEMINI_API_PRICING[model]
    except KeyError:
        # Unknown model
        return None

    # Calculate cost based on API type
    if use_vertex_ai:
        if model.startswith("gemini-2.0"):
            return _calculate_vertex_2_0_cost(metadata, model_pricing, model)
        elif model.startswith("gemini-1.5"):  # pragma: no cover
            return _calculate_vertex_1_5_cost(metadata, model_pricing, model)
    else:
        # Standard Gemini API pricing
        return _calculate_standard_gemini_cost(
            metadata, model_pricing, model, use_long_context
        )



from typing import cast

from mypy_boto3_bedrock_runtime.type_defs import (
    InferenceConfigurationTypeDef,
)

from ...base.call_params import CommonCallParams
from ..call_params import BedrockCallParams

BEDROCK_PARAM_MAPPING = {
    "temperature": "temperature",
    "max_tokens": "maxTokens",
    "top_p": "topP",
    "stop": "stopSequences",
}


def convert_common_call_params(common_params: CommonCallParams) -> BedrockCallParams:
    """Convert CommonCallParams to Bedrock parameters."""
    inference_config = {}

    for key, value in common_params.items():
        if key not in BEDROCK_PARAM_MAPPING or value is None:
            continue

        if key == "stop":
            inference_config["stopSequences"] = (
                [value] if isinstance(value, str) else value
            )
        else:
            inference_config[BEDROCK_PARAM_MAPPING[key]] = value

    if not inference_config:
        return cast(BedrockCallParams, {})

    return cast(
        BedrockCallParams,
        {"inferenceConfig": cast(InferenceConfigurationTypeDef, inference_config)},
    )



from typing import cast

from ...base.call_params import CommonCallParams
from ..call_params import AzureCallParams

AZURE_PARAM_MAPPING = {
    "temperature": "temperature",
    "max_tokens": "max_tokens",
    "top_p": "top_p",
    "frequency_penalty": "frequency_penalty",
    "presence_penalty": "presence_penalty",
    "seed": "seed",
    "stop": "stop",
}


def convert_common_call_params(common_params: CommonCallParams) -> AzureCallParams:
    """Convert CommonCallParams to Azure parameters."""
    return cast(
        AzureCallParams,
        {
            AZURE_PARAM_MAPPING[key]: [value] if isinstance(value, str) else value
            for key, value in common_params.items()
            if key in AZURE_PARAM_MAPPING and AZURE_PARAM_MAPPING[key] is not None
        },
    )



"""This module contains the type definition for the base call parameters."""

from typing import TypeVar

from typing_extensions import NotRequired, TypedDict


class BaseCallParams(TypedDict, total=False): ...  # pragma: no cover


_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)


class CommonCallParams(TypedDict, total=False):
    """Common parameters shared across LLM providers.

    Note: Each provider may handle these parameters differently or not support them at all.
    Please check provider-specific documentation for parameter support and behavior.

    Attributes:
        temperature: Controls randomness in the output (0.0 to 1.0).
        max_tokens: Maximum number of tokens to generate.
        top_p: Nucleus sampling parameter (0.0 to 1.0).
        frequency_penalty: Penalizes frequent tokens (-2.0 to 2.0).
        presence_penalty: Penalizes tokens based on presence (-2.0 to 2.0).
        seed: Random seed for reproducibility.
        stop: Stop sequence(s) to end generation.
    """

    temperature: NotRequired[float | None]
    max_tokens: NotRequired[int | None]
    top_p: NotRequired[float | None]
    frequency_penalty: NotRequired[float | None]
    presence_penalty: NotRequired[float | None]
    seed: NotRequired[int | None]
    stop: NotRequired[str | list[str] | None]



import base64
from abc import update_abstractmethods
from contextlib import asynccontextmanager
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from anyio import create_memory_object_stream
from mcp.client.session import ClientSession as MCPClientSession
from mcp.client.stdio import (
    StdioServerParameters,
    get_default_environment,
)
from mcp.types import (
    BlobResourceContents,
    CallToolResult,
    EmbeddedResource,
    ImageContent,
    JSONRPCMessage,
    Prompt,
    PromptMessage,
    Resource,
    TextContent,
    TextResourceContents,
    Tool,
)
from pydantic import AnyUrl, BaseModel

from mirascope.core import BaseTool
from mirascope.core.base import AudioPart, DocumentPart, ImagePart, TextPart
from mirascope.mcp._utils import (
    _UNSET,
    build_object_type,
    convert_prompt_message_to_base_message_params,
    create_tool_from_mcp_tool,
    json_schema_to_python_type,
    read_stream_exception_filer,
    snake_to_pascal,
)
from mirascope.mcp.client import MCPClient, sse_client, stdio_client


@pytest.mark.asyncio
async def test_stdio_client() -> None:
    """Test stdio_client context manager."""
    send_read, receive_read = create_memory_object_stream(10)
    send_write, _ = create_memory_object_stream(10)

    @asynccontextmanager
    async def mock_client(server_params):
        await send_read.send(
            JSONRPCMessage(jsonrpc="2.0", method="initialized", id="init")  # pyright: ignore [reportCallIssue]
        )
        await send_read.aclose()
        yield receive_read, send_write

    class MockClientSession:
        def __init__(
            self,
            read,
            write,
            read_timeout_seconds=None,
            sampling_callback=None,
            list_roots_callback=None,
        ):
            self.read = read
            self.write = write

        async def __aenter__(self):
            # Just consume messages until 'initialized' appears
            async for msg in self.read:  # pragma: no cover
                if msg.root.method == "initialized":
                    break
            return self

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            pass

        async def initialize(self):
            return

    with (
        patch("mirascope.mcp.client.mcp_stdio_client", new=mock_client),
        patch("mirascope.mcp.client.ClientSession", new=MockClientSession),
    ):
        params = StdioServerParameters(command="fake_cmd", args=[], env=None)
        async with stdio_client(params) as client:
            assert isinstance(client, MCPClient)

    send_read, receive_read = create_memory_object_stream(10)
    send_write, _ = create_memory_object_stream(10)

    with (
        patch("mirascope.mcp.client.mcp_sse_client", new=mock_client),
        patch("mirascope.mcp.client.ClientSession", new=MockClientSession),
    ):
        async with sse_client("http://example.com") as client:
            assert isinstance(client, MCPClient)


def test_get_default_environment():
    # Test get_default_environment from stdio.py
    env = get_default_environment()
    # We can't guarantee specific keys, but we ensure it's a dict
    assert isinstance(env, dict)


@pytest.mark.asyncio
async def test_read_stream_exception_filer():
    send, receive = create_memory_object_stream(10)
    expect_exception = ValueError("Invalid")
    mock = MagicMock()
    async with read_stream_exception_filer(receive, mock) as filtered, send:
        # send a normal message
        await send.send(JSONRPCMessage(jsonrpc="2.0", id="ok", method="ok"))  # pyright: ignore [reportCallIssue]

        # send an exception
        await send.send(expect_exception)
        # send another normal message
        await send.send(JSONRPCMessage(jsonrpc="2.0", id="another", method="another"))  # pyright: ignore [reportCallIssue]
        await send.aclose()

        messages = []
        async for msg in filtered:
            messages.append(msg)
        assert len(messages) == 2
        assert messages[0].root.id == "ok"
        assert messages[1].root.id == "another"
        mock.assert_called_once_with(expect_exception)


@pytest.mark.asyncio
async def test_read_stream_exception_filer_no_handler():
    send, receive = create_memory_object_stream(10)

    async with read_stream_exception_filer(receive) as filtered, send:
        # send a normal message
        await send.send(JSONRPCMessage(jsonrpc="2.0", id="ok", method="ok"))  # pyright: ignore [reportCallIssue]

        # send an exception
        await send.send(ValueError("Invalid"))
        # send another normal message
        await send.send(JSONRPCMessage(jsonrpc="2.0", id="another", method="another"))  # pyright: ignore [reportCallIssue]
        await send.aclose()

        messages = []
        async for msg in filtered:
            messages.append(msg)
        assert len(messages) == 2
        assert messages[0].root.id == "ok"
        assert messages[1].root.id == "another"


def test_snake_to_pascal():
    # Test snake_to_pascal coverage
    assert snake_to_pascal("hello_world") == "HelloWorld"


def test_json_schema_to_python_type_all_cases():
    # Already tested some cases, add a fallback test
    assert json_schema_to_python_type({"type": "integer"}) is int
    assert json_schema_to_python_type({"type": "unknown_type"}) is Any
    # object without properties
    assert isinstance(json_schema_to_python_type({"type": "object"}), type)


def test_build_object_type_full():
    # Test more complex schema
    properties = {
        "str_field": {"type": "string", "description": "str desc"},
        "num_field": {"type": "number"},
        "arr_field": {"type": "array", "items": {"type": "integer"}},
        "obj_field": {
            "type": "object",
            "properties": {"inner": {"type": "boolean"}},
            "required": ["inner"],
        },
    }
    required = ["str_field"]
    Model = build_object_type(properties, required, "ComplexModel")
    instance = Model(str_field="abc", obj_field={"inner": True})
    assert instance.str_field == "abc"  # pyright: ignore [reportAttributeAccessIssue]
    assert instance.obj_field.inner is True  # pyright: ignore [reportAttributeAccessIssue]
    assert instance.num_field is None  # pyright: ignore [reportAttributeAccessIssue]
    assert instance.arr_field is None  # pyright: ignore [reportAttributeAccessIssue]


def test_create_model_from_tool_extra():
    tool = Tool(
        name="extra_tool",
        inputSchema={
            "type": "object",
            "properties": {
                "opt_str": {"type": "string", "description": "optional string"},
                "req_arr": {"type": "array", "items": {"type": "integer"}},
                "nullable": {"type": "string", "nullable": True},
            },
            "required": ["req_arr"],
        },
    )

    Model = create_tool_from_mcp_tool(tool)
    Model.call = lambda self: None  # pyright: ignore [reportAttributeAccessIssue]
    update_abstractmethods(Model)
    instance = Model(req_arr=[1, 2, 3])  # pyright: ignore [reportCallIssue]
    assert instance.req_arr == [1, 2, 3]  # pyright: ignore [reportAttributeAccessIssue]
    assert instance.opt_str is _UNSET  # pyright: ignore [reportAttributeAccessIssue]


def test_convert_prompt_message_all_content_types():
    # Test coverage for all content branches
    # TextContent
    pm_text = PromptMessage(role="user", content=TextContent(type="text", text="hi"))
    res_text = convert_prompt_message_to_base_message_params(pm_text)
    assert res_text.content == "hi"

    # ImageContent
    img_data = base64.b64encode(b"fake_img").decode()
    pm_img = PromptMessage(
        role="assistant",
        content=ImageContent(type="image", data=img_data, mimeType="image/png"),
    )
    res_img = convert_prompt_message_to_base_message_params(pm_img)
    assert isinstance(res_img.content[0], ImagePart)
    assert res_img.content[0].image == b"fake_img"

    # EmbeddedResource with TextResourceContents
    pm_embed_text = PromptMessage(
        role="assistant",
        content=EmbeddedResource(
            type="resource",
            resource=TextResourceContents(
                uri=AnyUrl("http://example.com"), text="embedded txt"
            ),
        ),
    )
    res_embed_text = convert_prompt_message_to_base_message_params(pm_embed_text)
    assert res_embed_text.content == "embedded txt"

    # EmbeddedResource with BlobResourceContents - image
    img_blob = base64.b64encode(b"someimg").decode()
    pm_embed_img = PromptMessage(
        role="assistant",
        content=EmbeddedResource(
            type="resource",
            resource=BlobResourceContents(
                uri=AnyUrl("http://example.com/img.png"),
                blob=img_blob,
                mimeType="image/png",
            ),
        ),
    )
    res_embed_img = convert_prompt_message_to_base_message_params(pm_embed_img)
    assert isinstance(res_embed_img.content[0], ImagePart)
    assert res_embed_img.content[0].image == b"someimg"

    # EmbeddedResource with BlobResourceContents - pdf
    pdf_blob = base64.b64encode(b"%PDF-fake").decode()
    pm_embed_pdf = PromptMessage(
        role="assistant",
        content=EmbeddedResource(
            type="resource",
            resource=BlobResourceContents(
                uri=AnyUrl("http://example.com/doc.pdf"),
                blob=pdf_blob,
                mimeType="application/pdf",
            ),
        ),
    )
    res_embed_pdf = convert_prompt_message_to_base_message_params(pm_embed_pdf)
    assert isinstance(res_embed_pdf.content[0], DocumentPart)
    assert res_embed_pdf.content[0].document == b"%PDF-fake"

    # EmbeddedResource with audio
    audio_blob = base64.b64encode(b"FAKEAUDIO").decode()
    pm_embed_audio = PromptMessage(
        role="assistant",
        content=EmbeddedResource(
            type="resource",
            resource=BlobResourceContents(
                uri=AnyUrl("http://example.com/sound.wav"),
                blob=audio_blob,
                mimeType="audio/wav",
            ),
        ),
    )
    res_embed_audio = convert_prompt_message_to_base_message_params(pm_embed_audio)
    assert isinstance(res_embed_audio.content[0], AudioPart)
    assert res_embed_audio.content[0].audio == b"FAKEAUDIO"

    # Unsupported mime
    unsupp_blob = base64.b64encode(b"FAKE").decode()
    pm_unsupp = PromptMessage.model_construct(
        role="assistant",
        content=EmbeddedResource(
            type="resource",
            resource=BlobResourceContents(
                uri=AnyUrl("http://example.com/file.xyz"),
                blob=unsupp_blob,
                mimeType="application/xyz",
            ),
        ),
        validate=False,
    )
    with pytest.raises(ValueError):
        convert_prompt_message_to_base_message_params(pm_unsupp)

    # Empty mime
    unsupp_blob = base64.b64encode(b"FAKE").decode()
    pm_unsupp = PromptMessage.model_construct(
        role="assistant",
        content=EmbeddedResource(
            type="resource",
            resource=BlobResourceContents(
                uri=AnyUrl("http://example.com/file.xyz"),
                blob=unsupp_blob,
                mimeType=None,
            ),
        ),
        validate=False,
    )
    with pytest.raises(ValueError):
        convert_prompt_message_to_base_message_params(pm_unsupp)

    # Invalid role
    pm_invalid_role = PromptMessage.model_construct(
        role="other", content=TextContent(type="text", text="hi"), validate=False
    )
    with pytest.raises(ValueError):
        convert_prompt_message_to_base_message_params(pm_invalid_role)

    # Unsupported content type
    class FakeContent(BaseModel):
        pass

    pm_unsupported_content = PromptMessage.model_construct(
        role="assistant", content=FakeContent(), validate=False
    )
    with pytest.raises(ValueError):
        convert_prompt_message_to_base_message_params(pm_unsupported_content)


@pytest.mark.asyncio
async def test_mcp_client_methods():
    # Test MCPClient methods: list_resources, read_resource, list_prompts, get_prompt_template, list_tools
    mock_session = AsyncMock(spec=MCPClientSession)
    mock_session.list_resources.return_value.resources = [
        Resource(uri=AnyUrl("http://example.com"), name="res1")
    ]
    mock_session.read_resource.return_value.contents = [
        TextResourceContents(uri=AnyUrl("http://example.com"), text="Hello"),
        BlobResourceContents(
            uri=AnyUrl("http://example.com"), blob=base64.b64encode(b"FAKE").decode()
        ),
    ]
    mock_session.list_prompts.return_value.prompts = [Prompt(name="p1")]
    mock_message = PromptMessage(
        role="assistant", content=TextContent(type="text", text="Hello from prompt")
    )
    mock_session.get_prompt.return_value = type(
        "MockPromptResult", (), {"messages": [mock_message]}
    )
    mock_tool = Tool(
        name="test_tool",
        inputSchema={
            "type": "object",
            "properties": {"genre": {"type": "string"}},
            "required": ["genre"],
        },
    )
    mock_session.list_tools.return_value.tools = [mock_tool]
    mock_session.call_tool.return_value = CallToolResult(
        isError=False, content=[TextContent(type="text", text="Tool result")]
    )

    client = MCPClient(mock_session)

    # list_resources
    res = await client.list_resources()
    assert len(res) == 1
    assert res[0].name == "res1"

    # read_resource
    contents = await client.read_resource("http://example.com")
    assert contents == [
        TextPart(type="text", text="Hello"),
        BlobResourceContents(
            uri=AnyUrl("http://example.com/"), mimeType=None, blob="RkFLRQ=="
        ),
    ]

    # list_prompts
    prompts = await client.list_prompts()
    assert prompts[0].name == "p1"

    # get_prompt_template and async_prompt execution
    async_prompt = await client.get_prompt_template("p1")
    prompt_result = await async_prompt(genre="mystery")
    assert prompt_result[0].content == "Hello from prompt"

    # list_tools
    tools = await client.list_tools()
    assert len(tools) == 1
    t = tools[0]
    assert issubclass(t, BaseTool)

    base_tool = BaseTool.type_from_base_model_type(t)
    ret = await base_tool(genre="fantasy").call()  # pyright: ignore [reportCallIssue, reportAbstractUsage]
    assert ret == ["Tool result"]


@pytest.mark.asyncio
async def test_mcp_client_tool_with_image():
    # Test a tool returning ImageContent or EmbeddedResource
    mock_session = AsyncMock(spec=MCPClientSession)
    img_data = base64.b64encode(b"IMAGEDATA").decode()
    mock_tools = [
        Tool(
            name="img_tool",
            inputSchema={"type": "object", "properties": {}, "required": []},
        ),
        Tool(
            name="img_tool",
            inputSchema={"type": "object", "properties": {}, "required": []},
        ),
        Tool(
            name="error_tool",
            inputSchema={"type": "object", "properties": {}, "required": []},
        ),
    ]
    mock_session.list_tools.return_value.tools = mock_tools
    mock_session.call_tool.return_value = CallToolResult(
        isError=False,
        content=[ImageContent(type="image", data=img_data, mimeType="image/png")],
    )

    client = MCPClient(mock_session)
    tools = await client.list_tools()
    tool = tools[0]
    base_tool = BaseTool.type_from_base_model_type(tool)
    result = await base_tool().call()  # pyright: ignore [reportAbstractUsage,reportCallIssue]
    assert len(result) == 1
    assert isinstance(result[0], ImageContent)

    # Test EmbeddedResource directly
    blob_data = base64.b64encode(b"SOMEDATA").decode()
    mock_session.call_tool.return_value = CallToolResult(
        isError=False,
        content=[
            EmbeddedResource(
                type="resource",
                resource=BlobResourceContents(
                    uri=AnyUrl("https://example.com"),
                    blob=blob_data,
                    mimeType="application/octet-stream",
                ),
            )
        ],
    )
    result = await base_tool().call()  # pyright: ignore [reportAbstractUsage]
    # In this case call returns the content as is (EmbeddedResource)
    assert isinstance(result[0], EmbeddedResource)

    mock_session.call_tool.return_value = CallToolResult(
        isError=True,
        content=[],
    )

    error_tool = tools[1]
    error_tool_type = BaseTool.type_from_base_model_type(error_tool)
    with pytest.raises(RuntimeError):
        await error_tool_type().call()  # pyright: ignore [reportAbstractUsage]



import pytest

from mirascope.core.base.types import CostMetadata
from mirascope.core.costs import calculate_cost


def test_calculate_cost_with_none_metadata(monkeypatch):
    """Test that if metadata is None, a new CostMetadata is created with cached_tokens defaulting to 0."""

    def dummy_openai(metadata, model):
        assert isinstance(metadata, CostMetadata)
        assert metadata.cached_tokens == 0
        return 50.0

    monkeypatch.setitem(
        calculate_cost.__globals__, "openai_calculate_cost", dummy_openai
    )

    result = calculate_cost("openai", "gpt-3.5-turbo", None)
    assert result == 50.0


def test_calculate_cost_unknown_provider():
    """Test that an unsupported provider raises a ValueError."""
    metadata = CostMetadata(input_tokens=1, output_tokens=1, cached_tokens=0)
    with pytest.raises(ValueError):
        calculate_cost("unknown_provider", "model", metadata)  # pyright: ignore [reportArgumentType]


@pytest.mark.parametrize(
    "provider,dummy_func_name,expected_cost",
    [
        ("openai", "openai_calculate_cost", 42.0),
        ("anthropic", "anthropic_calculate_cost", 24.0),
        ("azure", "azure_calculate_cost", 12.0),
        ("bedrock", "bedrock_calculate_cost", 8.0),
        ("cohere", "cohere_calculate_cost", 15.0),
        ("gemini", "gemini_calculate_cost", 18.0),
        ("google", "google_calculate_cost", 20.0),
        ("groq", "groq_calculate_cost", 22.0),
        ("mistral", "mistral_calculate_cost", 25.0),
        ("vertex", "vertex_calculate_cost", 30.0),
        ("xai", "xai_calculate_cost", 35.0),
        ("litellm", "litellm_calculate_cost", 35.0),
    ],
)
def test_calculate_cost_providers(
    monkeypatch, provider, dummy_func_name, expected_cost
):
    """Test that calculate_cost correctly routes to the provider-specific cost function."""

    def dummy_func(metadata, model):
        return expected_cost

    monkeypatch.setitem(calculate_cost.__globals__, dummy_func_name, dummy_func)

    metadata = CostMetadata(input_tokens=5, output_tokens=5, cached_tokens=0)
    result = calculate_cost(provider, "dummy_model", metadata)
    assert result == expected_cost



"""The `GoogleStream` class for convenience around streaming LLM calls.

usage docs: learn/streams.md
"""

from typing import cast

from google.genai.types import (
    Candidate,
    Content,
    ContentDict,
    ContentListUnion,
    ContentListUnionDict,
    FinishReason,
    FunctionCall,
    GenerateContentResponse,
    GenerateContentResponseUsageMetadata,
    PartDict,
    Tool,
)

from ..base.stream import BaseStream
from ..base.types import CostMetadata
from .call_params import GoogleCallParams
from .call_response import GoogleCallResponse
from .call_response_chunk import GoogleCallResponseChunk
from .dynamic_config import GoogleDynamicConfig
from .tool import GoogleTool


class GoogleStream(
    BaseStream[
        GoogleCallResponse,
        GoogleCallResponseChunk,
        ContentDict,
        ContentDict,
        ContentDict,
        ContentListUnion | ContentListUnionDict,
        GoogleTool,
        Tool,
        GoogleDynamicConfig,
        GoogleCallParams,
        FinishReason,
    ]
):
    """A class for convenience around streaming Google LLM calls.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.google import google_call


    @google_call("google-1.5-flash", stream=True)
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"

    stream = recommend_book("fantasy")  # returns `GoogleStream` instance
    for chunk, _ in stream:
        print(chunk.content, end="", flush=True)
    ```
    """

    _provider = "google"

    def _construct_message_param(
        self, tool_calls: list[FunctionCall] | None = None, content: str | None = None
    ) -> ContentDict:
        """Constructs the message parameter for the assistant."""
        return {
            "role": "model",
            "parts": cast(
                list[PartDict],
                [{"text": content}]
                + (
                    [
                        {"function_call": tool_call}
                        for tool_call in (tool_calls or [])
                        if tool_calls
                    ]
                    or []
                ),
            ),
        }

    def construct_call_response(self) -> GoogleCallResponse:
        """Constructs the call response from a consumed GoogleStream.

        Raises:
            ValueError: if the stream has not yet been consumed.
        """
        if not hasattr(self, "message_param"):
            raise ValueError(
                "No stream response, check if the stream has been consumed."
            )
        candidates_token_count = (
            int(self.output_tokens) if self.output_tokens is not None else None
        )
        prompt_token_count = (
            int(self.input_tokens) if self.input_tokens is not None else None
        )
        total_token_count = int(candidates_token_count or 0) + int(
            prompt_token_count or 0
        )
        response = GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=self.finish_reasons[0]
                    if self.finish_reasons
                    else FinishReason.STOP,
                    content=Content(
                        role=self.message_param["role"],  # pyright: ignore [reportTypedDictNotRequiredAccess]
                        parts=self.message_param["parts"],  # pyright: ignore [reportTypedDictNotRequiredAccess, reportArgumentType]
                    ),
                )
            ],
            model_version=self.model,
            usage_metadata=GenerateContentResponseUsageMetadata(
                candidates_token_count=candidates_token_count,
                prompt_token_count=prompt_token_count,
                total_token_count=total_token_count,
            ),
        )

        return GoogleCallResponse(
            metadata=self.metadata,
            response=response,
            tool_types=self.tool_types,
            prompt_template=self.prompt_template,
            fn_args=self.fn_args if self.fn_args else {},
            dynamic_config=self.dynamic_config,
            messages=self.messages,
            call_params=self.call_params,
            call_kwargs=self.call_kwargs,
            user_message_param=self.user_message_param,
            start_time=self.start_time,
            end_time=self.end_time,
        )

    @property
    def cost_metadata(self) -> CostMetadata:
        return super().cost_metadata



"""Tests the `_groq_calculate_cost` function."""

from mirascope.core.base.types import CostMetadata, ImageMetadata
from mirascope.core.costs._groq_calculate_cost import calculate_cost


def test_calculate_cost() -> None:
    """Tests the `calculate_cost` function."""
    assert (
        calculate_cost(CostMetadata(), model="llama3-groq-70b-8192-tool-use-preview")
        is None
    )
    assert (
        calculate_cost(CostMetadata(input_tokens=1, output_tokens=1), model="unknown")
        is None
    )
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1, output_tokens=1),
            model="llama3-groq-70b-8192-tool-use-preview",
        )
        == 1.78e-6
    )


def test_calculate_cost_different_models() -> None:
    """Tests the `calculate_cost` function with different models."""
    # Test with mixtral model (default)
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1000, output_tokens=500),
        )
        == 0.00000024 * 1000 + 0.00000024 * 500
    )

    # Test with llama model
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1000, output_tokens=500),
            model="llama3-70b-8192",
        )
        == 0.00000059 * 1000 + 0.00000079 * 500
    )

    # Test with gemma model
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1000, output_tokens=500),
            model="gemma-7b-it",
        )
        == 0.00000007 * 1000 + 0.00000007 * 500
    )


def test_calculate_cost_vision_models() -> None:
    """Tests the `calculate_cost` function with vision models."""
    # Test vision model without images
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1000, output_tokens=500),
            model="llama-3.2-11b-vision-preview",
        )
        == 0.00000018 * 1000 + 0.00000018 * 500
    )

    # Test vision model with one image
    assert (
        calculate_cost(
            CostMetadata(
                input_tokens=1000,
                output_tokens=500,
                images=[ImageMetadata(width=0, height=0)],
            ),
            model="llama-3.2-11b-vision-preview",
        )
        == 0.00000018 * 1000 + 0.00000018 * 500 + 0.00000018 * 6400
    )

    # Test vision model with multiple images
    assert (
        calculate_cost(
            CostMetadata(
                input_tokens=1000,
                output_tokens=500,
                images=[
                    ImageMetadata(width=0, height=0),
                    ImageMetadata(width=0, height=0),
                    ImageMetadata(width=0, height=0),
                ],
            ),
            model="llama-3.2-90b-vision-preview",
        )
        == 0.01863
    )

    # Test with images on non-vision model (should ignore images)
    assert (
        calculate_cost(
            CostMetadata(
                input_tokens=1000,
                output_tokens=500,
                images=[ImageMetadata(width=0, height=0)],
            ),
            model="mixtral-8x7b-32768",
        )
        == 0.00000024 * 1000 + 0.00000024 * 500
    )


def test_calculate_cost_none_values() -> None:
    """Tests the `calculate_cost` function with None values."""
    # Test with None input_tokens
    assert (
        calculate_cost(
            CostMetadata(input_tokens=None, output_tokens=500),
            model="mixtral-8x7b-32768",
        )
        is None
    )

    # Test with None output_tokens
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1000, output_tokens=None),
            model="mixtral-8x7b-32768",
        )
        is None
    )


def test_calculate_cost_asymmetric_pricing() -> None:
    """Tests the `calculate_cost` function with models having asymmetric pricing."""
    # Test with model where input and output costs differ
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1000, output_tokens=500),
            model="llama-3.3-70b-specdec",
        )
        == 0.00000059 * 1000 + 0.00000099 * 500
    )

    # Test with deepseek model
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1000, output_tokens=500),
            model="deepseek-r1-distill-llama-70b",
        )
        == 0.00000075 * 1000 + 0.00000099 * 500
    )



"""This module contains the `fn_is_async` function."""

import inspect
from collections.abc import Awaitable, Callable, Coroutine
from typing import Any, TypeVar

from typing_extensions import TypeIs

_R = TypeVar("_R")


def fn_is_async(
    fn: Callable[..., Awaitable[_R] | Coroutine[Any, Any, _R]] | Callable[..., _R],
) -> TypeIs[Callable[..., Awaitable[_R]]]:
    if inspect.iscoroutinefunction(fn):
        return True

    # Check if it's a wrapper around an `async def` function (using functools.wraps).
    _fn = fn
    while hasattr(_fn, "__wrapped__"):
        _fn = _fn.__wrapped__  # pyright: ignore[reportFunctionMemberAccess]
        if inspect.iscoroutinefunction(_fn):
            return True
    return False



"""Tests the `google.call_response` module."""

from google.genai.types import (
    Candidate,
    Content,
    FunctionCall,
    GenerateContentResponse,
    Part,
)
from google.genai.types import FinishReason as GoogleFinishReason

from mirascope.core import BaseMessageParam
from mirascope.core.base.types import CostMetadata, GoogleMetadata
from mirascope.core.google.call_response import GoogleCallResponse
from mirascope.core.google.tool import GoogleTool


def test_google_call_response() -> None:
    """Tests the `GoogleCallResponse` class."""
    response = GenerateContentResponse(
        candidates=[
            Candidate(
                finish_reason=GoogleFinishReason.STOP,
                content=Content(
                    parts=[Part(text="The author is Patrick Rothfuss")],
                    role="model",
                ),
            )
        ]
    )

    call_response = GoogleCallResponse(
        metadata={},
        response=response,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    call_response._model = "google-1.5-flash"
    assert call_response._provider == "google"
    assert call_response.content == "The author is Patrick Rothfuss"
    assert call_response.finish_reasons == ["STOP"]
    assert call_response.model == "google-1.5-flash"
    assert call_response.id is None
    assert call_response.tools is None
    assert call_response.tool is None
    assert call_response.usage is None
    assert call_response.input_tokens is None
    assert call_response.output_tokens is None
    assert call_response.cost is None
    assert call_response.cost_metadata == CostMetadata(google=GoogleMetadata())
    assert call_response.cached_tokens is None
    assert call_response.message_param == {
        "role": "model",
        "parts": [Part(text="The author is Patrick Rothfuss")],
    }
    assert call_response.common_finish_reasons == ["stop"]
    assert call_response.common_message_param == BaseMessageParam(
        role="assistant",
        content="The author is Patrick Rothfuss",
    )


def test_google_call_response_with_tools() -> None:
    """Tests the `GoogleCallResponse` class with tools."""

    class FormatBook(GoogleTool):
        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    response = GenerateContentResponse(
        candidates=[
            Candidate(
                finish_reason=GoogleFinishReason.STOP,
                content=Content(
                    parts=[
                        Part(
                            function_call=FunctionCall(
                                name="FormatBook",
                                args={
                                    "title": "The Name of the Wind",
                                    "author": "Patrick Rothfuss",
                                },
                            )
                        ),
                    ],
                    role="model",
                ),
            )
        ]
    )

    call_response = GoogleCallResponse(
        metadata={},
        response=response,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    tools = call_response.tools
    tool = call_response.tool
    assert tools and len(tools) == 1 and tools[0] == tool
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    output = tool.call()
    assert output == "The Name of the Wind by Patrick Rothfuss"
    assert call_response.tool_message_params([(tool, output)]) == [
        {
            "parts": [
                {
                    "function_response": {
                        "name": "FormatBook",
                        "response": {
                            "result": "The Name of the Wind by Patrick Rothfuss"
                        },
                    }
                }
            ],
            "role": "user",
        }
    ]
    assert call_response.common_user_message_param is None
    call_response.user_message_param = {
        "role": "user",
        "parts": [{"text": "content"}],
    }
    assert call_response.common_user_message_param == BaseMessageParam(
        role="user", content="content"
    )
    call_response.messages = [
        {"role": "user", "parts": [{"text": "content"}]},
        {"role": "assistant", "parts": [{"text": "content"}]},
    ]
    assert call_response.common_messages == [
        BaseMessageParam(role="user", content="content"),
        BaseMessageParam(role="assistant", content="content"),
    ]



"""This module contains the setup_call function for OpenAI tools."""

import os
from collections.abc import Awaitable, Callable
from typing import Any, overload

from openai import AsyncOpenAI, OpenAI
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessageParam,
)
from pydantic import BaseModel

from ...base import BaseTool
from ...base._utils import AsyncCreateFn, CreateFn, fn_is_async
from ...base.call_params import CommonCallParams
from ...base.stream_config import StreamConfig
from ...openai import (
    AsyncOpenAIDynamicConfig,
    OpenAICallParams,
    OpenAIDynamicConfig,
    OpenAITool,
)
from ...openai._call_kwargs import OpenAICallKwargs
from ...openai._utils import setup_call as setup_call_openai


@overload
def setup_call(
    *,
    model: str,
    client: AsyncOpenAI | None,
    fn: Callable[..., Awaitable[AsyncOpenAIDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: AsyncOpenAIDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: OpenAICallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    AsyncCreateFn[ChatCompletion, ChatCompletionChunk],
    str | None,
    list[ChatCompletionMessageParam],
    list[type[OpenAITool]] | None,
    OpenAICallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: OpenAI | None,
    fn: Callable[..., OpenAIDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: OpenAIDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: OpenAICallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[ChatCompletion, ChatCompletionChunk],
    str | None,
    list[ChatCompletionMessageParam],
    list[type[OpenAITool]] | None,
    OpenAICallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: OpenAI | AsyncOpenAI | None,
    fn: Callable[..., OpenAIDynamicConfig]
    | Callable[..., Awaitable[AsyncOpenAIDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: OpenAIDynamicConfig | AsyncOpenAIDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: OpenAICallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[ChatCompletion, ChatCompletionChunk]
    | AsyncCreateFn[ChatCompletion, ChatCompletionChunk],
    str | None,
    list[ChatCompletionMessageParam],
    list[type[OpenAITool]] | None,
    OpenAICallKwargs,
]:
    if not client:
        api_key = os.environ.get("XAI_API_KEY")
        client = (
            AsyncOpenAI(api_key=api_key, base_url="https://api.x.ai/v1")
            if fn_is_async(fn)
            else OpenAI(api_key=api_key, base_url="https://api.x.ai/v1")
        )
    create, prompt_template, messages, tool_types, call_kwargs = setup_call_openai(
        model=model,  # pyright: ignore [reportCallIssue]
        client=client,  # pyright: ignore [reportArgumentType]
        fn=fn,  # pyright: ignore [reportArgumentType]
        fn_args=fn_args,  # pyright: ignore [reportArgumentType]
        dynamic_config=dynamic_config,
        tools=tools,
        json_mode=json_mode,
        call_params=call_params,
        response_model=response_model,
        stream=stream,
    )
    return create, prompt_template, messages, tool_types, call_kwargs



"""The Mirascope Gemini Module."""

import inspect
import warnings
from typing import TypeAlias

from google.generativeai.protos import FunctionResponse
from google.generativeai.types import ContentDict

from ..base import BaseMessageParam
from ._call import gemini_call
from ._call import gemini_call as call
from .call_params import GeminiCallParams
from .call_response import GeminiCallResponse
from .call_response_chunk import GeminiCallResponseChunk
from .dynamic_config import GeminiDynamicConfig
from .stream import GeminiStream
from .tool import GeminiTool

GeminiMessageParam: TypeAlias = ContentDict | FunctionResponse | BaseMessageParam

warnings.warn(
    inspect.cleandoc("""
    The `mirascope.core.gemini` module is deprecated and will be removed in a future release.
    Please use the `mirascope.core.google` module instead.
    """),
    category=DeprecationWarning,
)

__all__ = [
    "GeminiCallParams",
    "GeminiCallResponse",
    "GeminiCallResponseChunk",
    "GeminiDynamicConfig",
    "GeminiMessageParam",
    "GeminiStream",
    "GeminiTool",
    "call",
    "gemini_call",
]



"""This module contains the `BedrockCallResponseChunk` class.

usage docs: learn/streams.md#handling-streamed-responses
"""

from typing import cast

from pydantic import SkipValidation
from types_aiobotocore_bedrock_runtime.literals import StopReasonType as FinishReason

from ..base import BaseCallResponseChunk, types
from ..base.types import CostMetadata
from ._types import AsyncStreamOutputChunk, StreamOutputChunk, TokenUsageTypeDef
from ._utils._convert_finish_reason_to_common_finish_reasons import (
    _convert_finish_reasons_to_common_finish_reasons,
)


class BedrockCallResponseChunk(
    BaseCallResponseChunk[StreamOutputChunk | AsyncStreamOutputChunk, FinishReason]
):
    """A convenience wrapper around the Bedrock `ChatCompletionChunk` streamed chunks.

    When calling the Bedrock API using a function decorated with `bedrock_call` and
    `stream` set to `True`, the stream will contain `BedrockResponseChunk` instances with
    properties that allow for more convenient access to commonly used attributes.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.bedrock import bedrock_call


    @bedrock_call("anthropic.claude-3-haiku-20240307-v1:0", stream=True)
    @prompt_template("Recommend a {genre} book")
    def recommend_book(genre: str):
        ...


    stream = recommend_book("fantasy")  # response is an `BedrockStream`
    for chunk, _ in stream:
        print(chunk.content, end="", flush=True)
    ```
    """

    chunk: SkipValidation[StreamOutputChunk | AsyncStreamOutputChunk]

    @property
    def content(self) -> str:
        """Returns the content for the 0th choice delta."""
        if content_block_delta := self.chunk.get("contentBlockDelta"):
            return content_block_delta["delta"].get("text", "")
        return ""

    @property
    def finish_reasons(self) -> list[FinishReason]:
        """Returns the finish reasons of the response."""
        if (stop_reason := self.chunk.get("messageStop")) and stop_reason.get(
            "stopReason"
        ):
            return [stop_reason["stopReason"]]
        return []

    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return self.chunk["model"]

    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.chunk["responseMetadata"]["RequestId"]

    @property
    def usage(self) -> TokenUsageTypeDef | None:
        """Returns the usage of the chat completion."""
        usage = self.chunk.get("metadata", {}).get("usage")
        if usage:
            return usage
        return None

    @property
    def input_tokens(self) -> int | None:
        """Returns the number of input tokens."""
        if self.usage:
            return self.usage["inputTokens"]
        return None

    @property
    def cached_tokens(self) -> int | None:
        """Returns the number of cached tokens."""
        return None

    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        if self.usage:
            return self.usage["outputTokens"]
        return None

    @property
    def cost_metadata(self) -> CostMetadata:
        """Returns the cost metadata."""
        return super().cost_metadata

    @property
    def common_finish_reasons(self) -> list[types.FinishReason] | None:
        return _convert_finish_reasons_to_common_finish_reasons(
            cast(list[str], self.finish_reasons)
        )



"""Tests the `_utils.get_template_variables` module."""

from mirascope.core.base._utils._get_template_variables import get_template_variables


def test_get_template_variables() -> None:
    """Test the `get_template_variables` function."""
    template = "This is a {variable} template with {multiple:spec} variables."
    assert get_template_variables(template, False) == ["variable", "multiple"]
    assert get_template_variables(template, True) == [
        ("variable", ""),
        ("multiple", "spec"),
    ]



from typing import cast

from google.genai.types import GenerateContentConfig

from ...base.call_params import CommonCallParams
from ..call_params import GoogleCallParams

GOOGLE_PARAM_MAPPING = {
    "temperature": "temperature",
    "max_tokens": "max_output_tokens",
    "top_p": "top_p",
    "stop": "stop_sequences",
}


def convert_common_call_params(common_params: CommonCallParams) -> GoogleCallParams:
    """Convert CommonCallParams to Google parameters."""
    config = GenerateContentConfig()

    for key, value in common_params.items():
        if key not in GOOGLE_PARAM_MAPPING or value is None:
            continue

        if key == "stop":
            # Google API expects a list of strings for stop sequences
            if isinstance(value, str):
                config.stop_sequences = [value]
            elif isinstance(value, list):
                config.stop_sequences = value
        else:
            target_attr = GOOGLE_PARAM_MAPPING[key]
            setattr(config, target_attr, value)

    default_config = GenerateContentConfig()
    if config == default_config:
        return cast(GoogleCallParams, {})

    return cast(GoogleCallParams, {"config": config})



"""Handles the stream of completion chunks."""

import json
from collections.abc import AsyncGenerator, Generator

from mypy_boto3_bedrock_runtime.type_defs import (
    ToolUseBlockOutputTypeDef,
)
from typing_extensions import TypedDict

from .._types import (
    AsyncStreamOutputChunk,
    StreamOutputChunk,
    ToolUseBlockContentTypeDef,
)
from ..call_response_chunk import BedrockCallResponseChunk
from ..tool import BedrockTool


class ToolUseChunk(TypedDict):
    tool_use_id: str
    input_chunk: str
    name: str
    stop: bool


def _handle_chunk(
    chunk: StreamOutputChunk | AsyncStreamOutputChunk,
    current_tool_use_chunk: ToolUseChunk | None,
    tool_types: list[type[BedrockTool]] | None,
) -> tuple[
    BedrockCallResponseChunk | None,
    BedrockTool | None,
    ToolUseChunk | None,
]:
    """Handles a chunk of the stream."""
    if not tool_types:
        return BedrockCallResponseChunk(chunk=chunk), None, None
    elif (content_block_start := chunk.get("contentBlockStart")) and (
        tool_use := content_block_start["start"].get("toolUse")
    ):
        current_tool_use_chunk = ToolUseChunk(
            tool_use_id=tool_use["toolUseId"],
            input_chunk="",
            name=tool_use["name"],
            stop=False,
        )
    elif (
        (content_block_delta := chunk.get("contentBlockDelta"))
        and (tool_use := content_block_delta["delta"].get("toolUse"))
        and current_tool_use_chunk
        and not current_tool_use_chunk["stop"]
    ):
        current_tool_use_chunk["input_chunk"] += tool_use["input"]
        return None, None, current_tool_use_chunk
    elif "contentBlockStop" in chunk and current_tool_use_chunk:
        current_tool_use_chunk["stop"] = True
        return None, None, current_tool_use_chunk
    elif current_tool_use_chunk and current_tool_use_chunk["stop"]:
        for tool_type in tool_types:
            if current_tool_use_chunk["name"] == tool_type._name():
                current_tool_use = ToolUseBlockContentTypeDef(
                    toolUse=ToolUseBlockOutputTypeDef(
                        toolUseId=current_tool_use_chunk["tool_use_id"],
                        input=json.loads(current_tool_use_chunk["input_chunk"]),
                        name=current_tool_use_chunk["name"],
                    )
                )
                return (
                    BedrockCallResponseChunk(chunk=chunk),
                    tool_type.from_tool_call(current_tool_use),
                    None,
                )
    return BedrockCallResponseChunk(chunk=chunk), None, current_tool_use_chunk


def handle_stream(
    stream: Generator[StreamOutputChunk, None, None],
    tool_types: list[type[BedrockTool]] | None,
    partial_tools: bool = False,
) -> Generator[tuple[BedrockCallResponseChunk, BedrockTool | None], None, None]:
    """Iterator over the stream and constructs tools as they are streamed."""
    current_tool_use_chunk = None
    for chunk in stream:
        call_response, tool, current_tool_use_chunk = _handle_chunk(
            chunk, current_tool_use_chunk, tool_types
        )
        if call_response:
            yield call_response, tool


async def handle_stream_async(
    stream: AsyncGenerator[AsyncStreamOutputChunk, None],
    tool_types: list[type[BedrockTool]] | None,
    partial_tools: bool = False,
) -> AsyncGenerator[tuple[BedrockCallResponseChunk, BedrockTool | None], None]:
    """Async iterator over the stream and constructs tools as they are streamed."""
    current_tool_use_chunk = None
    async for chunk in stream:
        call_response, tool, current_tool_use_chunk = _handle_chunk(
            chunk, current_tool_use_chunk, tool_types
        )
        if call_response:
            yield call_response, tool



"""Tests the `stream` module."""

from functools import partial
from typing import cast
from unittest.mock import MagicMock, patch

import pytest

from mirascope.core.base.stream import BaseStream, stream_factory


@pytest.fixture()
def mock_stream_decorator_kwargs() -> dict:
    """Returns the mock kwargs (excluding fn) for the stream `decorator` function."""
    return {
        "model": "model",
        "tools": [],
        "json_mode": True,
        "client": MagicMock(),
        "call_params": MagicMock(),
    }


@patch("mirascope.core.base.stream.prompt_template", new_callable=MagicMock)
@patch("mirascope.core.base.stream.get_metadata", new_callable=MagicMock)
def test_stream_factory_sync(
    mock_get_metadata: MagicMock,
    mock_prompt_template_decorator: MagicMock,
    mock_setup_call: MagicMock,
    mock_stream_decorator_kwargs: dict,
) -> None:
    """Tests the `stream_factory` method on a sync function."""
    (
        mock_create,
        mock_prompt_template,
        mock_messages,
        mock_tool_types,
        mock_call_kwargs,
    ) = mock_setup_call.return_value
    mock_create = cast(MagicMock, mock_create)

    mock_handle_stream = MagicMock()
    mock_handle_stream.return_value = [("chunk", "tool")]
    mock_call_response_type = MagicMock
    decorator = partial(
        stream_factory(
            TCallResponse=mock_call_response_type,
            TStream=MagicMock,
            setup_call=mock_setup_call,
            handle_stream=mock_handle_stream,
            handle_stream_async=MagicMock(),
        ),
        **mock_stream_decorator_kwargs,
    )

    dynamic_config = MagicMock()
    mock_get = MagicMock()
    mock_get.return_value = None
    dynamic_config.get = mock_get

    def fn(genre: str, *, topic: str):
        """Recommend a {genre} book on {topic}."""
        return dynamic_config

    mock_prompt_template_decorator.return_value = lambda x: fn

    decorated_fn = decorator(fn)
    assert decorated_fn._model == mock_stream_decorator_kwargs["model"]  # pyright: ignore [reportFunctionMemberAccess]
    stream: BaseStream = decorated_fn(genre="fantasy", topic="magic")  # type: ignore
    assert list(stream.stream) == mock_handle_stream.return_value  # type: ignore

    assert stream.metadata == mock_get_metadata.return_value
    assert stream.tool_types == mock_tool_types
    assert stream.call_response_type == mock_call_response_type
    assert stream.model == mock_stream_decorator_kwargs["model"]
    assert stream.prompt_template == mock_prompt_template
    fn_args = {"genre": "fantasy", "topic": "magic"}
    assert stream.fn_args == fn_args
    assert stream.dynamic_config == dynamic_config
    assert stream.messages == mock_messages
    assert stream.call_params == mock_stream_decorator_kwargs["call_params"]
    assert stream.call_kwargs == mock_call_kwargs

    mock_setup_call.assert_called_once_with(
        model=mock_stream_decorator_kwargs["model"],
        client=mock_stream_decorator_kwargs["client"],
        fn=fn,
        fn_args=fn_args,
        dynamic_config=dynamic_config,
        tools=mock_stream_decorator_kwargs["tools"],
        json_mode=mock_stream_decorator_kwargs["json_mode"],
        call_params=mock_stream_decorator_kwargs["call_params"],
        response_model=None,
        stream=True,
    )
    mock_create.assert_called_once_with(stream=True, **mock_call_kwargs)


@patch("mirascope.core.base.stream.prompt_template", new_callable=MagicMock)
@patch("mirascope.core.base.stream.get_metadata", new_callable=MagicMock)
@pytest.mark.asyncio
async def test_stream_factory_async(
    mock_get_metadata: MagicMock,
    mock_prompt_template_decorator: MagicMock,
    mock_setup_call_async: MagicMock,
    mock_stream_decorator_kwargs: dict,
) -> None:
    """Tests the `stream_factory` method on an async function."""
    (
        mock_create,
        mock_prompt_template,
        mock_messages,
        mock_tool_types,
        mock_call_kwargs,
    ) = mock_setup_call_async.return_value
    mock_create = cast(MagicMock, mock_create)

    mock_handle_stream_async = MagicMock()
    mock_handle_stream_async.return_value.__aiter__.return_value = [("chunk", "tool")]
    mock_call_response_type = MagicMock
    decorator = partial(
        stream_factory(
            TCallResponse=mock_call_response_type,
            TStream=MagicMock,
            setup_call=mock_setup_call_async,
            handle_stream=MagicMock(),
            handle_stream_async=mock_handle_stream_async,
        ),
        **mock_stream_decorator_kwargs,
    )

    dynamic_config = MagicMock()
    mock_get = MagicMock()
    mock_get.return_value = None
    dynamic_config.get = mock_get

    async def fn(genre: str, *, topic: str):
        """Recommend a {genre} book on {topic}."""
        return dynamic_config

    mock_prompt_template_decorator.return_value = lambda x: fn

    stream: BaseStream = await decorator(fn)(genre="fantasy", topic="magic")  # type: ignore
    stream_response = []
    async for t in stream.stream:  # type: ignore
        stream_response.append(t)
    assert (
        stream_response == mock_handle_stream_async.return_value.__aiter__.return_value
    )  # type: ignore

    assert stream.metadata == mock_get_metadata.return_value
    assert stream.tool_types == mock_tool_types
    assert stream.call_response_type == mock_call_response_type
    assert stream.model == mock_stream_decorator_kwargs["model"]
    assert stream.prompt_template == mock_prompt_template
    fn_args = {"genre": "fantasy", "topic": "magic"}
    assert stream.fn_args == fn_args
    assert stream.dynamic_config == dynamic_config
    assert stream.messages == mock_messages
    assert stream.call_params == mock_stream_decorator_kwargs["call_params"]
    assert stream.call_kwargs == mock_call_kwargs

    mock_setup_call_async.assert_called_once_with(
        model=mock_stream_decorator_kwargs["model"],
        client=mock_stream_decorator_kwargs["client"],
        fn=fn,
        fn_args=fn_args,
        dynamic_config=dynamic_config,
        tools=mock_stream_decorator_kwargs["tools"],
        json_mode=mock_stream_decorator_kwargs["json_mode"],
        call_params=mock_stream_decorator_kwargs["call_params"],
        response_model=None,
        stream=True,
    )
    mock_create.assert_called_once_with(stream=True, **mock_call_kwargs)


@patch(
    "mirascope.core.base.stream.get_possible_user_message_param",
    new_callable=MagicMock,
)
@patch.multiple(BaseStream, __abstractmethods__=set())
@pytest.mark.asyncio
async def test_base_stream(mock_get_possible_user_message_param: MagicMock) -> None:
    """Tests the `BaseStream` class."""
    mock_construct_message_param = MagicMock()
    mock_construct_message_param.return_value = "mock_message_param"
    BaseStream._construct_message_param = mock_construct_message_param
    call_response_type = MagicMock
    mock_tool_message_params = MagicMock()
    call_response_type.tool_message_params = mock_tool_message_params

    mock_chunk = MagicMock()
    mock_chunk.content = "content"
    mock_chunk.input_tokens = 1
    mock_chunk.output_tokens = 2
    mock_chunk.model = "updated_model"

    mock_tool = MagicMock()
    mock_tool.tool_call = "tool_call"
    mock_tool.call = lambda: "tool_call_response"

    stream = BaseStream(
        stream=(t for t in [(mock_chunk, mock_tool)]),
        metadata={},
        tool_types=[],
        call_response_type=call_response_type,
        model="model",
        prompt_template="prompt_template",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
    )  # type: ignore
    assert (
        stream.user_message_param == mock_get_possible_user_message_param.return_value
    )
    assert list(stream) == [(mock_chunk, mock_tool)]
    mock_construct_message_param.assert_called_once_with(["tool_call"], "content")
    assert stream.message_param == "mock_message_param"
    assert stream.model == "updated_model"

    async def generator():
        yield mock_chunk, mock_tool

    stream.stream, stream.message_param = generator(), None
    stream_response, tools_and_outputs = [], []
    async for chunk, tool in stream:
        stream_response.append((chunk, tool))
        tools_and_outputs.append((tool, tool.call()))  # type: ignore
    assert stream_response == [(mock_chunk, mock_tool)]
    mock_construct_message_param.assert_called_with(["tool_call"], "content")
    assert stream.message_param == "mock_message_param"
    assert stream.model == "updated_model"

    assert stream.tool_message_params(tools_and_outputs)
    mock_tool_message_params.assert_called_once_with(tools_and_outputs)



"""This module contains the type definition for the OpenAI call keyword arguments."""

from collections.abc import Sequence

from openai.types.chat import ChatCompletionMessageParam, ChatCompletionToolParam

from ..base import BaseCallKwargs
from .call_params import OpenAICallParams


class OpenAICallKwargs(OpenAICallParams, BaseCallKwargs[ChatCompletionToolParam]):
    model: str
    messages: Sequence[ChatCompletionMessageParam]



import warnings
from contextlib import suppress

from . import base

with suppress(ImportError):
    from . import anthropic

with suppress(ImportError):
    from . import openai


warnings.warn("""MIRASCOPE LOOK-ALIKE V0 INTERFACE USAGE
    You are using the `mirascope.v0` module, which implements look-alike interfaces.
    It is extremely important to note the following about these interfaces:
        - We provide no guarantee that we will support these interfaces. This includes any additional features or bugfixes.
        - We reserve the right to remove these interfaces from the library entirely.
        - These interfaces are an attempt at replicating the v0 interfaces, but they are built using v1 under the hood.
        - This means that you can access certain v1 features through these interfaces (such as multimodal prompt templates).
        - However, this also means that there are inevitable differences (e.g. `OpenAICallResponse` returned from `call` is the `v1` implementation).
    
    The goal of these interfaces is to make it possible to migrate from v0 to v1 without requiring a full migration.
    HOWEVER, there are a few key differences to note:
        - STREAMING:
            - Was annoying to replicate, so we have instead opted to support streaming with the v1 return typing.
            - The return type is a tuple[Chunk, Tool], so if you're not using tools, simply ignore the second item (e.g. `for chunk, _ in stream: ...`)
            - If you're using tools, simply access tools as they are streamed through the second tuple item (i.e. no more `OpenAIToolStream`)
        - BASE CONFIG:
            - Was also annoying to replicate, so we've opted not to replicate this.
            - To provide a custom client (e.g. `AzureOpenAI`), simply pass that client into your function (e.g. `book_recommender.call(client=AzureOpenAI(...))`) and we'll use it.
            - This means there is no longer a `BaseConfig` object on anything.
        - We currently only support the v0 look-alike interfaces for Anthropic and OpenAI.
        - The import structure may differ slightly. In particular, there are no submodules for `anthropic` or `openai`, so you should be importing directly from those modules (e.g. `from mirascope.v0.openai import OpenAICall`)
              
    Next steps:
        - We encourage you to try out the new v1 interface and mirgrate over. We think it's a big improvement, and we hope you will come to feel this way too.
              - We've written a migration guide to help with this process: https://mirascope.com/MIGRATE
        - We greatly appreciate any and all questions/comments/feedback, so if you have anything, send it our way!
        - We understand that some people may prefer the class approach from v0 interface design. You can achieve a very similar interface using v1, see https://mirascope.com/MIGRATE/#baseprompt
        - If there turns out to be sufficient desire for the class-based interfaces to be included in v1 as fully supported interfaces, we will potentially consider this.
""")

__all__ = ["anthropic", "base", "openai"]



"""Calculate the cost of a completion using the Groq API."""

from ..base.types import CostMetadata


def calculate_cost(
    metadata: CostMetadata,
    model: str = "mixtral-8x7b-32768",
) -> float | None:
    """Calculate the cost of a completion using the Groq API.

    https://wow.groq.com/

    Model                                  Input                  Output
    llama-3.3-70b-versatile                $0.59 / 1M tokens      $0.79 / 1M tokens
    llama-3.3-70b-specdec                  $0.59 / 1M tokens      $0.99 / 1M tokens
    llama-3.1-8b-instant                   $0.05 / 1M tokens      $0.08 / 1M tokens
    llama3-70b-8192                        $0.59 / 1M tokens      $0.79 / 1M tokens
    llama3-8b-8192                         $0.05 / 1M tokens      $0.08 / 1M tokens
    llama-guard-3-8b                       $0.20 / 1M tokens      $0.20 / 1M tokens
    mixtral-8x7b-32768                     $0.24 / 1M tokens      $0.24 / 1M tokens
    gemma-7b-it                            $0.07 / 1M tokens      $0.07 / 1M tokens
    gemma2-9b-it                           $0.20 / 1M tokens      $0.20 / 1M tokens
    mistral-saba-24b                       $0.79 / 1M tokens      $0.79 / 1M tokens
    qwen-2.5-32b                           $0.79 / 1M tokens      $0.79 / 1M tokens
    qwen-2.5-coder-32b                     $0.79 / 1M tokens      $0.79 / 1M tokens
    deepseek-r1-distill-qwen-32b           $0.69 / 1M tokens      $0.69 / 1M tokens
    deepseek-r1-distill-llama-70b          $0.75 / 1M tokens      $0.99 / 1M tokens
    deepseek-r1-distill-llama-70b-specdec  $0.75 / 1M tokens      $0.99 / 1M tokens
    llama-3.2-1b-preview                   $0.04 / 1M tokens      $0.04 / 1M tokens
    llama-3.2-3b-preview                   $0.06 / 1M tokens      $0.06 / 1M tokens
    llama-3.2-11b-vision-preview           $0.18 / 1M tokens      $0.18 / 1M tokens
    llama-3.2-90b-vision-preview           $0.90 / 1M tokens      $0.90 / 1M tokens
    """
    pricing = {
        "llama-3.3-70b-versatile": {
            "prompt": 0.000_000_59,
            "completion": 0.000_000_79,
        },
        "llama-3.3-70b-specdec": {
            "prompt": 0.000_000_59,
            "completion": 0.000_000_99,
        },
        "llama3-groq-70b-8192-tool-use-preview": {
            "prompt": 0.000_000_89,
            "completion": 0.000_000_89,
        },
        "llama3-groq-8b-8192-tool-use-preview": {
            "prompt": 0.000_000_19,
            "completion": 0.000_000_19,
        },
        "llama-3.1-8b-instant": {
            "prompt": 0.000_000_05,
            "completion": 0.000_000_08,
        },
        "llama-guard-3-8b": {
            "prompt": 0.000_000_2,
            "completion": 0.000_000_2,
        },
        "llama3-70b-8192": {
            "prompt": 0.000_000_59,
            "completion": 0.000_000_79,
        },
        "llama3-8b-8192": {
            "prompt": 0.000_000_05,
            "completion": 0.000_000_08,
        },
        "mixtral-8x7b-32768": {
            "prompt": 0.000_000_24,
            "completion": 0.000_000_24,
        },
        "gemma-7b-it": {
            "prompt": 0.000_000_07,
            "completion": 0.000_000_07,
        },
        "gemma2-9b-it": {
            "prompt": 0.000_000_2,
            "completion": 0.000_000_2,
        },
        "mistral-saba-24b": {
            "prompt": 0.000_000_79,
            "completion": 0.000_000_79,
        },
        "qwen-2.5-32b": {
            "prompt": 0.000_000_79,
            "completion": 0.000_000_79,
        },
        "qwen-2.5-coder-32b": {
            "prompt": 0.000_000_79,
            "completion": 0.000_000_79,
        },
        "deepseek-r1-distill-qwen-32b": {
            "prompt": 0.000_000_69,
            "completion": 0.000_000_69,
        },
        "deepseek-r1-distill-llama-70b": {
            "prompt": 0.000_000_75,
            "completion": 0.000_000_99,
        },
        "deepseek-r1-distill-llama-70b-specdec": {
            "prompt": 0.000_000_75,
            "completion": 0.000_000_99,
        },
        "llama-3.2-1b-preview": {
            "prompt": 0.000_000_04,
            "completion": 0.000_000_04,
        },
        "llama-3.2-3b-preview": {
            "prompt": 0.000_000_06,
            "completion": 0.000_000_06,
        },
        # Vision models
        "llama-3.2-11b-vision-preview": {
            "prompt": 0.000_000_18,
            "completion": 0.000_000_18,
        },
        "llama-3.2-90b-vision-preview": {
            "prompt": 0.000_000_90,
            "completion": 0.000_000_90,
        },
    }

    if metadata.input_tokens is None or metadata.output_tokens is None:
        return None

    try:
        model_pricing = pricing[model]
    except KeyError:
        return None

    # Calculate cost for text tokens
    prompt_cost = metadata.input_tokens * model_pricing["prompt"]
    completion_cost = metadata.output_tokens * model_pricing["completion"]
    total_cost = prompt_cost + completion_cost

    # Calculate cost for images if present
    # Groq bills 6,400 tokens per image for vision models
    # https://groq.com/pricing/
    image_cost = 0.0
    if metadata.images and "vision" in model:
        # For Groq vision models, each image is billed at 6,400 tokens
        tokens_per_image = 6400

        # Count the number of images
        image_count = len(metadata.images)

        # Calculate total image tokens
        total_image_tokens = image_count * tokens_per_image

        # Images are charged at the prompt token rate
        image_cost = total_image_tokens * model_pricing["prompt"]

    # Add image cost to total
    total_cost += image_cost

    return total_cost



from typing_extensions import TypedDict


class StreamConfig(TypedDict):
    """Configuration options for streaming.

    Attributes:
        partial_tools (bool): Whether to stream partial tool responses
    """

    partial_tools: bool



"""Calculate the cost of a completion using the Mistral API."""

from ..base.types import CostMetadata


def calculate_cost(
    metadata: CostMetadata,
    model: str = "open-mistral-7b",
) -> float | None:
    """Calculate the cost of a completion using the Mistral API.

    https://mistral.ai/technology/#pricing

    Model                     Input               Cached     Output
    mistral-large-latest      $2/1M tokens                   $6/1M tokens
    pixtral-large-latest      $2/1M tokens                   $6/1M tokens
    mistral-small-latest      $0.1/1M tokens                 $0.3/1M tokens
    mistral-saba-latest       $0.2/1M tokens                 $0.6/1M tokens
    codestral-latest          $0.3/1M tokens                 $0.9/1M tokens
    ministral-8b-latest       $0.1/1M tokens                 $0.1/1M tokens
    ministral-3b-latest       $0.04/1M tokens                $0.04/1M tokens
    mistral-embed             $0.1/1M tokens                 -
    mistral-moderation-latest $0.1/1M tokens                 -
    open-mistral-nemo         $0.3/1M tokens                 $0.3/1M tokens
    open-mistral-7b           $0.25/1M tokens                $0.25/1M tokens
    open-mixtral-8x7b         $0.7/1M tokens                 $0.7/1M tokens
    open-mixtral-8x22b        $2/1M tokens                   $6/1M tokens
    """
    pricing = {
        "mistral-large-latest": {"prompt": 0.000_002, "completion": 0.000_006},
        "pixtral-large-latest": {"prompt": 0.000_002, "completion": 0.000_006},
        "mistral-small-latest": {"prompt": 0.000_000_1, "completion": 0.000_000_3},
        "mistral-saba-latest": {"prompt": 0.000_000_2, "completion": 0.000_000_6},
        "codestral-latest": {"prompt": 0.000_000_3, "completion": 0.000_000_9},
        "ministral-8b-latest": {"prompt": 0.000_000_1, "completion": 0.000_000_1},
        "ministral-3b-latest": {"prompt": 0.000_000_04, "completion": 0.000_000_04},
        "mistral-embed": {"prompt": 0.000_000_1, "completion": 0},
        "mistral-moderation-latest": {"prompt": 0.000_000_1, "completion": 0},
        "open-mistral-nemo": {"prompt": 0.000_000_3, "completion": 0.000_000_3},
        "open-mistral-nemo-2407": {"prompt": 0.000_000_3, "completion": 0.000_000_3},
        "open-mistral-7b": {"prompt": 0.000_000_25, "completion": 0.000_000_25},
        "open-mixtral-8x7b": {"prompt": 0.000_000_7, "completion": 0.000_000_7},
        "open-mixtral-8x22b": {"prompt": 0.000_002, "completion": 0.000_006},
        "mistral-large-2407": {"prompt": 0.000_003, "completion": 0.000_009},
        "mistral-medium-latest": {"prompt": 0.000_002_75, "completion": 0.000_008_1},
        "pixtral-12b-2409": {"prompt": 0.000_002, "completion": 0.000_006},
    }

    if metadata.input_tokens is None or metadata.output_tokens is None:
        return None

    try:
        model_pricing = pricing[model]
    except KeyError:
        return None

    # Calculate cost for text tokens
    prompt_cost = metadata.input_tokens * model_pricing["prompt"]
    completion_cost = metadata.output_tokens * model_pricing["completion"]
    total_cost = prompt_cost + completion_cost

    # Image tokens is included in the cost

    return total_cost
